# -*- coding: utf-8 -*-
"""
header
======

Module to convert parsed data from an Amira (R) header into a set of nested objects.
The key class is :py:class:`ahds.header.AmiraHeader`.

Usage:

::

    >>> from ahds.header import AmiraHeader
    >>> ah = AmiraHeader('<somfile>.am')
    >>> print(ah)
    +-header                                                                                       AmiraHeader [is_parent? True ]
    |  +-filetype: AmiraMesh
    |  +-dimension: None
    |  +-format: BINARY
    |  +-endian: LITTLE
    |  +-version: 2.1
    |  +-extra_format: None
    |  +-Parameters                                                                                      Block [is_parent? True ]
    |  |  +-Materials                                                                                ListBlock [is_parent? True ]
    |  |  |  +[0]-Exterior                                                                               Block [is_parent? False]
    |  |  |  |  +-Id: 1
    |  |  |  +[1]-Inside                                                                                 Block [is_parent? False]
    |  |  |  |  +-Color: [0.64, 0, 0.8]
    |  |  |  |  +-Id: 2
    |  |  |  +[2]-Mitochondria                                                                           Block [is_parent? False]
    |  |  |  |  +-Id: 3
    |  |  |  |  +-Color: [0, 1, 0]
    |  |  |  +[3]-Mitochondria_                                                                          Block [is_parent? False]
    |  |  |  |  +-Id: 4
    |  |  |  |  +-Color: [1, 1, 0]
    |  |  |  +[4]-mitochondria__                                                                         Block [is_parent? False]
    |  |  |  |  +-Id: 5
    |  |  |  |  +-Color: [0, 0.125, 1]
    |  |  |  +[5]-NE                                                                                     Block [is_parent? False]
    |  |  |  |  +-Id: 6
    |  |  |  |  +-Color: [1, 0, 0]
    |  |  +-Content: 862x971x200 byte, uniform coordinates
    |  |  +-BoundingBox: [0, 13410.7, 0, 15108.4, 1121.45, 4221.01]
    |  |  +-CoordType: uniform
    |  +-Lattice                                                                                         Block [is_parent? False]
    |  |  +-length: [862 971 200]


Each nested object is constructed from the :py:class:``Block`` class defined in the `core` module.

There are five top-level attributes that every ``AmiraHeader`` will have:

* `designation`
* `definitions`
* `Parameters`
* `Materials`
* `data_pointers`

All of them, except `Materials` and `Parameters` are marked deprecated and may be removed in
future. The `designation` and `defintions` attributes just return the reference to the `AmiraHeader`
instance which host all the attributes and data array declarations (`definitions`) found within
the `AmiraMesh` and `HyperSurface` files. The `data_pointers` attribute collects from each
attribute representing a valid array declaration the defined data definitions and returns the resulting
list. New code should directly access the Block objects describing the available data directly
from the corresponding array attribute of the `AmiraHeader` object.

The blocks defining the available data structures extend the basic `Block` to the corresponding
`AmiraMeshDataStream` and `AmiraHxSurfaceDataStream` blocks which allow to load and access the correspinding
data using the `data` property defined by theses blocks. The additional `stream_data`, `encoded_data` and
`decoded_data` attributes are deprecated and may be removed in future versions.

The `attrs` attribute of all `Block` and subclass objects including `AmiraHeader` itself
which allows to query the attributes added during loading is deprecated and may be removed in future version.
To query which attributes are defined on each block use the builtin dir function instead. To
access the defined attributes just use them liḱe any other static attribute.

The following attributes are moved from the `designation` attribute to the basic `header` block:

* `filetype` e.g. ``AmiraMesh``, ``Avizo`` or ``HyperSurface``
* `dimension` e.g. ``3D``
* `format` e.g. ``BINARY-LITTLE-ENDIAN``
* `version` e.g. ``2.1``
* `extra_format`

Data pointers are identified by the name ``data_pointer_<n>`` where <n> is the number indicated by the
block attibute in the above examples.

The above approach can also be used to load Amira HyperSurface files.

    >>> ah = AmiraHeader('<somfile>.surf')
    +-header                                                                                       AmiraHeader [is_parent? False]
    |  +-filetype: HyperSurface
    |  +-dimension: None
    |  +-format: BINARY
    |  +-endian: BIG
    |  +-version: 0.1
    |  +-extra_format: None
    |  +-Parameters                                                                                      Block [is_parent? True ]
    |  |  +-Materials                                                                                ListBlock [is_parent? True ]
    |  |  |  +[0]-Exterior                                                                               Block [is_parent? False]
    |  |  |  |  +-Id: 1
    |  |  |  +[1]-Background                                                                             Block [is_parent? False]
    |  |  |  |  +-Id: 1
    |  |  |  |  +-Color: [1, 0, 0]
    |  |  |  +[2]-host_cell_2                                                                            Block [is_parent? False]
    |  |  |  |  +-Id: 2
    |  |  |  |  +-Color: [1, 1, 0.167]
    |  |  +-BoundaryIds                                                                                  Block [is_parent? False]
    |  |  |  +-Name: BoundaryConditions
    |  |  +-GridBox: [-1, 850, -1, 932, -1, 233]
    |  |  +-GridSize: [852, 934, 235]
    |  |  +-Filename: /Users/ubanan01/Desktop/cryo_Tomo_Polara/U2OS/08.04....

NOTE!!!
NEW DESCRIPTION OF THE FILE FORMAT:
https://assets.thermofisher.com/TFS-Assets/MSD/Product-Guides/user-guide-amira-software.pdf

"""
from __future__ import print_function

import sys
import numpy
import warnings

from .core import Block, deprecated, ListBlock
from .data_stream import set_data_stream
from .grammar import get_parsed_data


class AmiraHeader(Block):
    """Class to encapsulate Amira metadata and accessors to Amira (R) data streams"""

    # __slots__ field is used to separate internal attributes from the dynamic attributes
    # representing the metadata and the accessors to the content of the data streams
    # which will be stored inside the __dict__ attribute of the Block base class
    __slots__ = (
        '_fn', '_parsed_data', '_header_length', '_file_format', '_parameters', '_load_streams',
        '_data_stream_count')

    # fixme: load_streams should be False by default
    def __init__(self, fn, load_streams=True, *args, **kwargs):
        """Construct an AmiraHeader object from parsed data"""
        self._fn = fn
        self._literal_data, self._parsed_data, self._header_length, self._file_format = get_parsed_data(fn, *args,
                                                                                                        **kwargs)
        # load the streams
        self._load_streams = load_streams
        # data stream count
        self._data_stream_count = None
        super(AmiraHeader, self).__init__('header')
        # load the parse data into this object
        self._load()

    @classmethod
    @deprecated("Now you can directly create a header from the file name as AmiraHeader('file.am')")
    def from_file(cls, fn, *args, **kwargs):
        """Deprecated classmethod"""
        return cls(fn, *args, **kwargs)

    @property
    def filename(self):
        return self._fn

    @property
    def literal_data(self):
        return self._literal_data

    @property
    def parsed_data(self):
        return self._parsed_data

    # todo: change this to streams_loaded
    @property
    def load_streams(self):
        return self._load_streams

    # todo: change this to streams_loaded
    @load_streams.setter
    def load_streams(self, value):
        try:
            assert isinstance(value, bool)
        except AssertionError:
            raise TypeError("must be a bool")
        self._load_streams = value

    @property
    def data_stream_count(self):
        return self._data_stream_count

    def load(self):
        """Public loading method"""
        self._load()

    def __len__(self):
        return self._header_length

    @staticmethod
    def flatten_dict(in_dict):
        block_data = dict()
        for block in in_dict:
            for block_keys0 in block.keys():
                block_data[block_keys0] = block[block_keys0]
                break
        return block_data

    def _load(self):
        # first flatten the dict
        block_data = self.flatten_dict(self._parsed_data)
        # load file designations
        self._load_designation(block_data['designation'])
        # load parameters
        if 'parameters' in block_data:
            _parameters = self._load_parameters(block_data['parameters'], 'Parameters', parent=self)
        else:
            # just create an empty parameters block to keep header consistent
            _parameters = Block('Parameters')
        # if we have a Materials block in parameters we create a convenience dictionary for
        # accessing materials e.g. for patches
        if hasattr(_parameters, 'Materials'):
            material_dict = dict()
            for material in _parameters.Materials:
                material_dict[material.name] = material
            _parameters.Materials.material_dict = material_dict
        super(AmiraHeader, self).add_attr('Parameters', _parameters)
        # load array declarations
        self._load_declarations(block_data['array_declarations'])
        # load data stream definitions
        if self.filetype == "AmiraMesh":
            # a list of data streams
            self._data_streams_block_list = self._load_definitions(block_data['data_definitions'])
            self._data_stream_count = len(self._data_streams_block_list)
        elif self.filetype == "HyperSurface":
            # data_streams_list = self._locate_hx_streams()
            self._data_streams_block_list = []
            self._data_stream_count = 1
            if self.load_streams:
                block = set_data_stream('Data', self)
                block.read()

    @deprecated(" use header attributes version, dimension, fileformat, format and extra_format istead")
    def designation(self):
        """Designation of the Amira file defined in the first row
        
        Designations consist of some or all of the following data:
        
        * filetype e.g. ``AmiraMesh`` or ``HyperSurface``
        * dimensions e.g. ``3D``
        * format e.g. ``BINARY-LITTLE-ENDIAN``
        * version e.g. ``2.1``
        * extra format e.g. ``<hxsurface>``

        NOTE: this property is deprecated use the corresponding attributes of the AmiraHeader
        instead to access the above informations
        """
        return self

    @deprecated(" use data array attributes instead eg. Vertices, Triangles, ...")
    def definitions(self):
        """Definitions consist of a key-value pair specified just after the 
        designation preceded by the key-word 'define'      

        NOTE: this property is deprecated access the corresponding attributes directly
        eg. ah.Nodes instead of ah.definitions.Nodes or ah.Tetrahedra instead of ah.defintions.Tetrahedra
        """
        return self

    @deprecated(" use data attributes of data arrays  instead eg. header.Vertices.Coordinates")
    def data_pointers(self):
        """The list of data pointers together with a name, data type, dimension, 
        index, format and length

        NOTE: deprecated access the data defnitions for each data array through the corresponding attributes
        eg.: ah.Nodes.Coordinates instead of ah.data_pointers.data_pointer_1 ah.Tetrahedra.Nodes instead of 
        ah.data_pointers.data_pointer_2 etc.
        """

        _data_pointers = Block("data_pointers")
        return _data_pointers

    def _load_designation(self, block_data):
        self.add_attr('filetype', block_data.get('filetype', None))
        self.add_attr('dimension', block_data.get('dimension', None))
        format = block_data.get('format')
        if format == 'BINARY':
            self.add_attr('format', format)
            self.add_attr('endian', 'BIG')
        elif format == 'ASCII':
            self.add_attr('format', format)
            self.add_attr('endian', None)
        elif format == 'BINARY-LITTLE-ENDIAN':
            self.add_attr('format', 'BINARY')
            self.add_attr('endian', 'LITTLE')
        else:
            raise ValueError(
                u'unsupported format {format}; kindly consider contacting the maintainer to include support. Thanks.'.format(
                    format=format))
        self.add_attr('version', block_data.get('version', None))
        self.add_attr('extra_format', block_data.get('extra_format', None))

    def _load_parameters(self, block_data, name='Parameters', **kwargs):
        # treat materials specially (and possibly others in the future)
        if name in ["Materials"]:
            block = ListBlock(name)
            for param in block_data:
                # a sequence of parameters
                if isinstance(param['parameter_value'], list):
                    if len(param['parameter_value']) > 0:
                        if param['parameter_value'][0] == '<!?c?!>':
                            block.add_attr(param['parameter_name'], param['parameter_value'][1:])
                        else:
                            block.append(
                                self._load_parameters(
                                    param['parameter_value'],
                                    name=param['parameter_name']
                                )
                            )
                    else:
                        block.add_attr(param['parameter_name'], param['parameter_value'])
                # a string or number
                else:
                    block.add_attr(param['parameter_name'], param['parameter_value'])
        else:
            block = Block(name)
            for param in block_data:
                try:
                    if isinstance(param['parameter_value'], list):
                        if len(param['parameter_value']) > 0:
                            if param['parameter_value'][0] == '<!?c?!>':
                                block.add_attr(param['parameter_name'], param['parameter_value'][1:])
                            else:
                                block.add_attr(
                                    self._load_parameters(param['parameter_value'], name=param['parameter_name'],parent_block=block))
                        else:
                            # print(param['parameter_name'], type(param['parameter_name']))
                            block.add_attr(param['parameter_name'], param['parameter_value'])
                    else:
                        parameter_name = param['parameter_name']
                        if parameter_name == 'name':
                            parameter_value = param['parameter_value']
                            if parameter_value != name:
                                parent_block = kwargs.get('parent_block',None)
                                if isinstance(parent_block,Block):
                                    parent_block.add_attr(parameter_value,block)
                                else: # may be removed to silently ignore alias definitions on Parameters
                                    warnings.warn("Setting alias '{}' for block '{}' failed: No parent".format(parameter_value,name))
                                block.add_attr('@alias',parameter_value)
                        else:
                            block.add_attr(param['parameter_name'], param['parameter_value'])
                except KeyError:
                    print(u"Found odd parameter: {} = {}".format(list(param.keys())[0], param[list(param.keys())[0]]),
                          file=sys.stderr)
        return block

    def _load_declarations(self, block_data):
        """Load the array definition blocks which will contain the data streams"""
        for decl in block_data:
            block = Block(decl['array_name'])
            block.add_attr('length', decl['array_dimension'])
            self.add_attr(block)

    def _load_definitions(self, block_data):
        """We want to load data definitions to the appropriate array definition block"""
        data_streams = list()
        data_stream_indices = set()
        for defn in block_data:
            # check whether the array_ref is an attribute on self
            try:
                parent = getattr(self, defn['array_reference'])
            except AttributeError:
                parent = self
            # set the data streams
            block = set_data_stream(defn['data_name'], self)
            block.add_attr('data_index', defn['data_index'])
            # conditionally add the data stream if its index is unique e.g. Fields do not have unique ds
            if defn['data_index'] not in data_stream_indices:
                data_streams.append(block)
            # keep track of the data stream indices
            data_stream_indices.add(defn['data_index'])
            block.add_attr('dimension', defn.get('data_dimension', 1))  # assume dimension of 1
            block.add_attr('type', defn['data_type'])
            block.add_attr('interpolation_method', defn.get('interpolation_method', None))
            _shape = getattr(parent, 'length', None)
            if isinstance(_shape, numpy.ndarray):
                block.add_attr('shape', tuple(_shape.tolist()[::-1]))
            else:
                block.add_attr('shape', _shape)
            block.add_attr('format', defn.get('data_format', None))
            # insert this definition as an attribute
            # parent.add_attr(block)
            # keep track of data streams
        return data_streams

    def __repr__(self):
        return "AmiraHeader('{}')".format(self.filename)




# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import unittest

import numpy

import ahds
from ahds import data_stream, AmiraFile, header
from ahds.tests import TEST_DATA_PATH


class TestDataStreams(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.af_am = AmiraFile(os.path.join(TEST_DATA_PATH, 'test9.am'))
        cls.af_am_hxsurf = AmiraFile(os.path.join(TEST_DATA_PATH, 'test8.am'))
        cls.af_hxsurf = AmiraFile(os.path.join(TEST_DATA_PATH, 'test7.surf'))

    def test_amirafile(self):
        self.assertIsInstance(self.af_am.header, header.AmiraHeader)
        self.assertIsInstance(self.af_am, AmiraFile)
        self.assertIsInstance(self.af_am_hxsurf, AmiraFile)
        self.assertIsInstance(self.af_hxsurf, AmiraFile)

    def test_data_stream(self):
        # af_am_data_stream = self.af_am.data_streams[1]
        # test9.am
        self.assertIsInstance(self.af_am.data_streams, ahds.Block)
        self.assertEqual(self.af_am.header.data_stream_count, 1)
        self.assertEqual(self.af_am.data_streams.name, 'data_streams')
        self.assertEqual(self.af_am.data_streams.Labels.name, 'Labels')

        # test8.am
        self.assertIsInstance(self.af_am_hxsurf.data_streams, ahds.Block)
        self.assertEqual(self.af_am_hxsurf.header.data_stream_count, 3)
        self.assertEqual(self.af_am_hxsurf.data_streams.name, 'data_streams')
        self.assertEqual(self.af_am_hxsurf.data_streams.Vertices.name, 'Vertices')
        self.assertEqual(self.af_am_hxsurf.data_streams.Triangles.name, 'Triangles')
        self.assertEqual(getattr(self.af_am_hxsurf.data_streams, "Patches-0").name, "Patches-0")

        # test7.surf
        self.assertIsInstance(self.af_hxsurf.data_streams, ahds.Block)
        self.assertEqual(self.af_hxsurf.header.data_stream_count, 1)
        self.assertEqual(self.af_hxsurf.data_streams.name, 'data_streams')
        self.assertEqual(self.af_hxsurf.data_streams.Data.name, 'Data')


        # self.assertIsInstance(af_am_data_stream, data_stream.AmiraMeshDataStream)
        # self.assertEqual(len(self.af_am.data_streams), 1)
        # self.assertEqual(len(self.af_am_hxsurf.data_streams), 3)
        # self.assertEqual(len(self.af_hxsurf.data_streams), 5)
        # self.assertIsInstance(self.af_hxsurf.data_streams['Patches'], data_stream.PatchesDataStream)

        # self.assertTrue(len(af_am_data_stream.encoded_data) > 0)
        # self.assertTrue(len(af_am_data_stream.decoded_data) > 0)

    # def test_images(self):
    #     imgs = self.af_am.data_streams[1].to_images()
    #     self.assertTrue(len(imgs) > 0)
    #
    # def test_images_fail(self):
    #     with self.assertRaises(ValueError):
    #         self.af_am_hxsurf.data_streams[1].to_images()

    # def test_volume(self):
    #     vol = self.af_am.data_streams[1].to_volume()
    #     self.assertIsInstance(vol, numpy.ndarray)
    #     x, y, z = vol.shape
    #     self.assertTrue(x > 0)
    #     self.assertTrue(y > 0)
    #     self.assertTrue(z > 0)

    # def test_contours(self):
    #     imgs = self.af_am.data_streams[1].to_images()
    #     # get the middle slice of the image set
    #     contours = imgs[128].as_contours
    #     self.assertIsInstance(contours, dict)



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import random
import sys
import unittest

from . import TEST_DATA_PATH, Py23FixTestCase
from .. import AmiraFile
from ..core import Block, ListBlock, _print


# class TestUtils(unittest.TestCase):
#     pass


class TestBlock(Py23FixTestCase):
    def test_create(self):
        b = Block('block')
        self.assertTrue(b.name, 'block')
        self.assertFalse(b.is_parent)
        with self.assertRaises(AttributeError):
            b.name = 'something else'

    def test_add_attrs(self):
        b = Block('block')
        i = Block('inner-block')
        print(dir(b))
        print(dir(i))
        b.add_attr(i)
        b.add_attr('user', 'oranges')
        self.assertTrue(b.is_parent)
        self.assertFalse(i.is_parent)
        self.assertListEqual(b.attrs(), ['inner-block', 'user'])
        self.assertTrue(hasattr(b, 'inner-block'))
        self.assertTrue(hasattr(b, 'user'))
        with self.assertRaises(ValueError):
            b.add_attr(i)
        b.add_attr(Block('Materials'))
        inside = Block('Inside')
        b.Materials.add_attr(inside)
        # b.Materials.Inside.add_attr('Id', 1)
        outside = Block('Outside')
        b.Materials.add_attr(outside)
        # b.Materials.Outside.add_attr('Id', 2)
        seaside = Block('Seaside')
        b.Materials.add_attr(seaside)
        # b.Materials.Seaside.add_attr('Id', 3)
        # self.assertCountEqual(b.Materials.ids, [1, 2, 3])
        # self.assertEqual(b.Materials[1], inside)
        # self.assertEqual(b.Materials[2], outside)
        # self.assertEqual(b.Materials[3], seaside)
        self.assertTrue('Materials' in b)

    def test_rename_attr(self):
        b = Block('block')
        i = Block('inner-block')
        b.add_attr(i)
        self.assertEqual(getattr(b, 'inner-block'), i)
        b.rename_attr('inner-block', 'inner_block')
        self.assertEqual(b.inner_block, i)
        self.assertFalse(hasattr(b, 'inner-block'))
        # now retry to see if we can overwrite
        b = Block('block')
        b.add_attr(Block('inner-block'))
        b.add_attr(Block('inner_block'))
        with self.assertRaises(ValueError):
            b.rename_attr('inner-block', 'inner_block')
        with self.assertRaises(AttributeError):
            b.rename_attr('_inner-block', 'a')


class TestListBlock(Py23FixTestCase):
    def test_create(self):
        l = ListBlock('listblock')
        self.assertTrue(hasattr(l, '_list'))

    def test_add_attr(self):
        b = Block('block')
        b.add_attr(ListBlock('Materials'))
        inside = Block('Inside')
        b.Materials.append(inside)
        # b.Materials.add_attr(inside)
        b.Materials[0].add_attr('Id', 1)
        outside = Block('Outside')
        b.Materials.append(outside)
        b.Materials[1].add_attr('Id', 2)
        seaside = Block('Seaside')
        seaside.add_attr('Id', 3)
        b.Materials.append(seaside)
        self.assertCountEqual(b.Materials.ids, [1, 2, 3])
        self.assertEqual(b.Materials[0], inside)
        self.assertEqual(b.Materials[1], outside)
        self.assertEqual(b.Materials[2], seaside)
        self.assertTrue('Materials' in b)

    def test_list(self):
        l = ListBlock('listblock')
        l.append(Block('one'))
        l.append(Block('two'))
        # l[0] = Block('one')
        # l[1] = Block('two')
        l[0].add_attr(Block('inside'))
        l[0].add_attr('value', 10)
        self.assertEqual(len(l), 2)
        l.add_attr('internal', 33)
        l.internal
        self.assertTrue(hasattr(l, 'internal'))

    def test_parentage(self):
        l = ListBlock('A new listblock')
        self.assertFalse(l.is_parent)
        l.add_attr('volex', 'Lufthansa')
        self.assertFalse(l.is_parent)  # is only a parent if the list is populated or if it has subblocks
        l.add_attr(Block('inner-block'))
        self.assertTrue(l.is_parent)  # is a parent because the super class is a parent
        k = ListBlock('The other shoe')
        self.assertFalse(k.is_parent)
        k.append(Block('together'))
        # k[0] = Block('together')
        self.assertTrue(k.is_parent)

    def test_errors(self):
        l = ListBlock('test')
        # try to add a non-block as a list
        with self.assertRaises(ValueError):
            l[0] = 'yes'
        # try to get a non-existent value from the list
        with self.assertRaises(IndexError):
            l[1]
        # try to delete non-existent
        with self.assertRaises(IndexError):
            del l[1]

    def test_list_methods(self):
        """Mutable sequences should provide methods append(), count(), index(), extend(), insert(), pop(), remove(), reverse() and sort(), like Python standard list objects."""
        # append
        l = ListBlock('listblock')
        block = Block('nothing')
        l.append(block)
        self.assertEqual(len(l), 1)
        # fails for non-block
        with self.assertRaises(ValueError):
            l.append(1)
        # count
        l.append(block)
        self.assertEqual(l.count(block), 2)
        # index
        self.assertEqual(l.index(block), 0)
        # extend
        l.extend([Block('new-item')])
        self.assertEqual(len(l), 3)
        # insert
        l.insert(1, Block('inserted'))
        self.assertEqual(l[1].name, 'inserted')
        # pop
        l.pop()
        self.assertEqual(len(l), 3)
        # remove
        l.remove(block)
        self.assertEqual(l.count(block), 1)
        self.assertEqual(len(l), 2)
        # reverse
        l.append(Block('true'))
        l.append(Block('false'))
        items = list(l)
        l.reverse()
        ritems = list(l)
        self.assertListEqual(items[::-1], ritems)
        # sort
        l.insert(0, Block('zarathusa'))
        self.assertFalse(l[0] < l[1])
        l.sort()
        self.assertTrue(l[0] < l[1])

    # def test_insert_listblock(self):
    #     pass


class BlockSubclass(Block):
    """Subclass for test"""
    orange = 'pink'

    def __init__(self, *args, **kwargs):
        super(BlockSubclass, self).__init__(*args, **kwargs)
        self.init_prop = 'init_prop'

    @property
    def prop(self):
        return 'prop'


class ListBlockSubclass(ListBlock):
    """Subclass for test"""

    def __init__(self, *args, **kwargs):
        super(ListBlockSubclass, self).__init__(*args, **kwargs)

    def __str__(self, prefix="", index=None):
        return super(ListBlockSubclass, self).__str__(prefix=prefix, index=index)


class HxSurfSegment(object):
    """Generic HxSurface segment class

    The `ahds <http://ahds.readthedocs.io/en/latest/>`_ package provides a better abstraction of this filetype
    """

    def __init__(self, material, vertices, triangles, prune=True):
        """A single segment from an Amira HxSurface file

        Each such segment corresponds to an HxSurface patch. This is a convenience class to present
        key attributes (id, name, colour).

        :param material: an individual Block object with the Material block
        :param dict vertices: a dictionary of vertices indexed by vertex_ids (which appear in the triangles list)
        :param list triangles: a list of 3-lists containing vertex IDs which define each triangle
        :param bool prune: whether or not the prune the vertices for this segment/patch so that we only bear the
        vertices referenced here; default is True
        """
        self._material = material
        # id
        self._segment_id = self._material.Id
        # name
        if self._material.name:
            self._name = self._material.name
        else:
            self._name = None
        # colour
        if self._material.Color:
            self._colour = self._material.Color
        else:
            r, g, b = random.random(), random.random(), random.random()
            self._colour = r, g, b
            print("Warning: random colour ({:.4f}, {:.4f}, {:.4f}) for segment {}".format(r, g, b, self._segment_id),
                  file=sys.stderr)
        # vertices and triangles
        if prune:
            self._vertices = self._prune_vertices(vertices, triangles)
        else:
            self._vertices = vertices
        self._triangles = triangles
        # print(self._triangles[:20], list(map(lambda x: (x, self._vertices[x]), list(self._vertices.keys())[:20])))

    @property
    def id(self):
        """The segment ID"""
        return self._segment_id

    @property
    def name(self):
        """The name of the segment"""
        return self._name

    @property
    def colour(self):
        """The colour of the segment"""
        return self._colour

    @property
    def vertices(self):
        """A dictionary of vertices in this segment indexed by vertex ID"""
        return self._vertices

    @property
    def triangles(self):
        """A list of triangles (lists with 3 vertex IDs) in this segment"""
        return self._triangles

    def _prune_vertices(self, vertices, triangles):
        """Reduce the vertices and triangles to only those required by this segment"""
        # flatten the list of vertex ids in triangles
        unique_vertex_ids = set([vertex for triangle in triangles for vertex in triangle])
        # get only those vertices present in this segments triangles
        unique_vertices = {vertex: vertices[vertex] for vertex in unique_vertex_ids}
        return unique_vertices


def extract_segments(af, *args, **kwargs):
    """Extract patches as segments

    :param af: an `AmiraFile` object
    :return dict segments: a dictionary of segments with keys set to Material Ids (voxel values)
    """
    # make sure it's an AmiraFile object
    try:
        assert isinstance(af, AmiraFile)
    except AssertionError:
        raise TypeError("must be a valid AmiraFile object")
    # make sure it's read otherwise read it
    if not af.meta.streams_loaded:
        print("Data streams not yet loaded. Reading...", file=sys.stderr)
        af.read()
    segments = dict()
    # first we make a dictionary of vertices
    # keys are indices (1-based)
    vertices_list = af.data_streams.Data.Vertices.data
    # a dictionary of all vertices
    vertices_dict = dict(zip(range(1, len(vertices_list) + 1), vertices_list))
    # then we repack the vertices and patches into vertices and triangles (collate triangles from all patches)
    for patch in af.data_streams.Data.Vertices.Patches:
        material = af.header.Parameters.Materials.material_dict[patch.InnerRegion]
        patch_id = material.Id
        # sanity check
        if patch_id is None:
            raise ValueError('patch ID is None')
        # now collate triangles and vertices
        triangles = patch.Triangles.data
        hxsurfsegment = HxSurfSegment(material, vertices_dict, triangles.tolist(), *args, **kwargs)
        if patch_id not in segments:
            segments[patch_id] = [hxsurfsegment]
        else:
            segments[patch_id] += [hxsurfsegment]
    return segments


class TestBlockSubclass(unittest.TestCase):
    def test_block(self):
        """Test for attribute conflicts

        - properties
        - class attributes
        - instance attributes
        - custom attributes (retrieved using __getattr__)
        """
        b = BlockSubclass('b')
        b.add_attr('nothing', 0)
        self.assertTrue(hasattr(b, 'prop'))
        self.assertTrue(hasattr(b, 'init_prop'))
        self.assertTrue(hasattr(b, 'nothing'))
        self.assertTrue(hasattr(b.__class__, 'orange'))
        self.assertTrue(hasattr(b, 'orange'))
        with self.assertRaises(ValueError):
            b.add_attr('prop', None)
        with self.assertRaises(ValueError):
            b.add_attr('init_prop')
        with self.assertRaises(ValueError):
            b.add_attr('orange')
        # all other attributes are not affected
        with self.assertRaises(ValueError):
            b._attrs = 'attrs'
        with self.assertRaises(ValueError):
            b._attrs = list()
        # we can set an empty dictionary though
        b._attrs = dict()
        self.assertEqual(len(b._attrs), 0)
        # we can set a dictionary with string keys and anything as values
        _dict = {
            'a': 1,
            'very': 2,
            'big': Block('blocker'),
            'orangutang': 'that is what we hoped for'
        }
        b._attrs = _dict
        print(b)
        self.assertEqual(b._attrs, _dict)
        # check that we catch wrong dicts
        _bad_dict = {
            1: 1,
            2: 2,
            'big': Block('blocker'),
            'orangutang': 'that is what we hoped for'
        }
        with self.assertRaises(ValueError):
            b._attrs = _bad_dict

    def test_listblock(self):
        """Test that we cannot interfere with _list attribute"""
        l = ListBlockSubclass('l')
        _list = "a new list with spaces".split(' ')
        with self.assertRaises(ValueError):
            l._list = _list
        l._list = list(map(lambda x: Block(x), _list))
        self.assertEqual(len(l), len(_list))
        # test we can set an empty list
        l._list = []
        self.assertEqual(len(l), 0)


class TestAmiraFile(unittest.TestCase):
    """Tests for the ahds.AmiraFile class"""

    def test_default(self):
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'BinaryCustomLandmarks.elm'), load_streams=True)
        print(af)
        print()
        print()
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'testscalar.am'), load_streams=True)
        print(af)
        print()
        print()
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'test7.surf'), load_streams=True)
        print(af)

    """
    af = ahds.AmiraFile(fn, *args, **kwargs)
    header = af.header

    # TODO: handle <hxsurface> from .am files
    if header.designation.extra_format == "<hxsurface>":
        return header, None
    else:
        # read now
        af.read()
        data_streams = af.data_streams

        if len(data_streams) == 1:
            # get the index for the first (and only) data pointer
            index = header.data_pointers.data_pointer_1.data_index
            volume = data_streams[index].to_volume()
            return header, volume
        else:
            # get the first one and warn the user
            print_date("Multiple lattices defined. Is this file formatted properly? Trying to work with the first one...")
            index = header.data_pointers.data_pointer_1.data_index
            volume = data_streams[index].to_volume()
            return header, volume
    """

    def test_amreader_amiramesh(self):
        """Starting point for integrating into sfftk"""
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'test9.am'), load_streams=False)
        # print(af)
        print(af.header)
        # print(af.data_streams)
        self.assertEqual(af.header.data_stream_count, 1)
        # print(data_streams.Lattice.data.shape)
        af.read()
        print(af.data_streams)
        # print(dir(af))
        # print(af.attrs())
        # print(af.data_streams.attrs())
        data = getattr(af.data_streams, af.data_streams.attrs()[0], None)
        self.assertIsNotNone(data)
        # print(data.attrs())
        self.assertEqual(data.data_index, 1)
        _print(data.shape, type(data.shape))
        _print(af.header.Lattice.length, type(af.header.Lattice.length))
        self.assertEqual(data.shape, tuple(af.header.Lattice.length.tolist()))
        print(data.shape)

    def test_amreader_hxsurface(self):
        """Test that it correctly handles AmirMesh hxsurf files"""
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'test8.am'), load_streams=True)
        self.assertEqual(af.header.extra_format, "<hxsurface>")

    def test_amreader_amiramesh_multiple_volumes(self):
        """We assume a segmentation has one volume so..."""
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'TetraMesh.am'), load_streams=True)
        self.assertTrue(af.header.data_stream_count > 1)
        print(af.data_streams)

    """
    header = ahds.header.AmiraHeader.from_file(fn, *args, **kwargs)
    data_streams = ahds.data_stream.DataStreams(fn, *args, **kwargs)
    segments = dict()
    for patch_name in data_streams['Patches']:
        patch_material = getattr(header.Parameters.Materials, patch_name)
        patch_vertices, patch_triangles = vertices_for_patches(data_streams['Vertices'], data_streams['Patches'][patch_name])
        # we use the material ID as the key because it is a unique reference to the patch
        segments[patch_material.Id] = HxSurfSegment(patch_material, patch_vertices, patch_triangles)
    return header, segments
    """

    def test_surfreader(self):
        """The way sfftk uses ahds"""
        # af = AmiraFile(os.path.join(TEST_DATA_PATH, 'test7.surf'), load_streams=True)
        # af = AmiraFile(os.path.join(TEST_DATA_PATH, 'BinaryHyperSurface.surf'), load_streams=True)
        # af = AmiraFile('/Users/pkorir/data/segmentations/surf/test8.surf')
        af = AmiraFile(os.path.join(TEST_DATA_PATH, 'test7.surf'))
        # print(af.header.attrs())
        # print(af.header.Parameters.Materials)
        print(af.data_streams.Data.Vertices)
        with self.assertRaises(TypeError):
            extract_segments(list())
        segments = extract_segments(af)
        # check the structure of segments
        self.assertIsInstance(segments, dict)
        self.assertIsInstance(list(segments.values())[0], list)
        self.assertIsInstance(list(segments.keys())[0], int)
        self.assertIsInstance(list(segments.values())[0][0], HxSurfSegment)
        # if we have Materials verify that material_dict is not None
        if hasattr(af.header.Parameters, 'Materials'):
            self.assertIsNotNone(af.header.Parameters.Materials.material_dict)
        else:
            self.assertIsNone(af.header.Parameters.Materials.material_dict)
        # print(segments)
        # print(list(map(lambda s: (s.id, s.name, s.colour, len(s.vertices), len(s.triangles)), segments.values())))
        # print(segments[patch_id].id)
        # print(segments[patch_id].name)
        # print(segments[patch_id].colour)
        # print(len(segments[patch_id].vertices))  # , len(segments[2].vertices))
        # print(len(segments[patch_id].triangles))  # , len(segments[2].triangles))
        # self.assertEqual(segments[patch_id].id, )
        # for patch_id, hxsurfsegments in segments.items():
        #     print("""patch_id: {}""".format(patch_id))
        #     for s in hxsurfsegments:
        #         print(
        #             """Id:{}
        #             \rName:{}
        #             \rColour:{}
        #             \rNo. vertices: {}
        #             \rNo. triangles: {}
        #             """.format(s.id, s.name, s.colour, len(s.vertices), len(s.triangles)))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
ahds
====

This module provides a simple entry-point for using the underlying functionality
through the `AmiraFile` class which automatically handles both `AmiraMesh` and
`HxSurface` files. An `AmiraFile` is also a `Block` subclass with special attributes
`meta` - for metadata not explicitly provided in the file (such as `header_length`),
`header` - for the parse header and `data_streams` with the actual data stream data.

The only required argument is the name of the file to be read. By default, data streams
are loaded but can be turned off (for quick reading) by
setting `load_stream=False`. Additional `kwargs` are passed to the `AmiraHeader` class
call.

There is a `read` method which (if data streams have not yet been read) will read
the data streams.

An `AmiraFile` object may be printed to view the hierarchy of entities above or
passed to `repr` to view the instatiation call that represents it.

"""

import sys

from .core import Block
from .data_stream import set_data_stream
from .header import AmiraHeader

if sys.version_info[0] > 2:
    from shutil import get_terminal_size

    _get_terminal_size = get_terminal_size
else:
    from backports.shutil_get_terminal_size import get_terminal_size

    _get_terminal_size = get_terminal_size

WIDTH = _get_terminal_size().columns


class AmiraFile(Block):
    """Main entry point for working with Amira files"""
    __slots__ = ('_fn', '_load_streams', '_meta' '_header', '_data_streams')

    def __init__(self, fn, load_streams=True, *args, **kwargs):
        """Initialise a new AmiraFile object given the Amira file.

        Passes additional args/kwargs to AmiraHeader class for initialisation of the reading process

        .. code:: python

            from ahds import AmiraFile
            af = AmiraFile('file.am')
            print(af)

        :param str fn: Amira file name
        :param bool load_streams: whether (default) or not to load data streams
        """
        super(AmiraFile, self).__init__(fn)
        self._fn = fn
        self._load_streams = load_streams
        self._streams_loaded = False
        # the header contains a lot of information relied on for reading streams
        self._header = AmiraHeader(fn, load_streams=load_streams, *args, **kwargs)
        # meta block
        super(AmiraFile, self).add_attr('meta', Block('meta'))
        self.meta.add_attr('file', self._fn)
        self.meta.add_attr('header_length', len(self._header))
        self.meta.add_attr('data_streams', self._header.data_stream_count)
        self.meta.add_attr('streams_loaded', self._load_streams)
        # header block
        super(AmiraFile, self).add_attr('header', self._header)
        # data streams block
        super(AmiraFile, self).add_attr(Block('data_streams'))
        if self._load_streams:
            self.read()
            self._streams_loaded = True

    def read(self):
        """Read the data streams if they are not read yet"""
        if not self._streams_loaded:
            if self._header.filetype == "AmiraMesh":
                for ds in self._header._data_streams_block_list:
                    ds.read()
                    ds.add_attr('data', ds.get_data())
                    self.data_streams.add_attr(ds)
            elif self._header.filetype == "HyperSurface":
                block = set_data_stream('Data', self._header)
                block.read()
                self.data_streams.add_attr(block)
            self._load_streams = self._header.load_streams = True
            self._streams_loaded = True

    def __repr__(self):
        return "AmiraFile('{}', read={})".format(self._fn, self._read)

    def __str__(self, prefix="", index=None,alt_name=None):
        width = 140
        string = ""
        string += '*' * width + '\n'
        string += "AMIRA (R) HEADER AND DATA STREAMS\n"
        string += "-" * width + "\n"
        string += super(AmiraFile, self).__str__(prefix=prefix, index=index,alt_name=alt_name)
        string += "*" * width
        return string


__all__ = ['AmiraFile', 'AmiraHeader']



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import unittest

from ahds import grammar
from ahds.tests import TEST_DATA_PATH


class TestGrammar(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.file_format = grammar.detect_format(os.path.join(TEST_DATA_PATH, 'testscalar.am'))
        cls.header = grammar.get_header(os.path.join(TEST_DATA_PATH, 'testscalar.am'), cls.file_format)
        cls.parsed_header = grammar.parse_header(cls.header)

    def test_detect_format(self):
        self.assertIn(self.file_format, ['AmiraMesh', 'HyperSurface'])

    def test_get_header(self):
        self.assertTrue(len(self.header) > 0)

    def test_parse_header(self):
        self.assertTrue(len(self.parsed_header) > 0)





# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sys

from ahds import header
from ahds.tests import Py23FixTestCase, TEST_DATA_PATH


class TestHeaderBlock(Py23FixTestCase):
    @classmethod
    def setUpClass(cls):
        cls.header = header.AmiraHeader(os.path.join(TEST_DATA_PATH, 'test12.am'))

    def test_amiraheader(self):
        self.assertIsInstance(self.header, header.AmiraHeader)

    def test_add_attr(self):
        self.header.Parameters.add_attr('x', 10)
        self.assertTrue(hasattr(self.header.Parameters, 'x'))
        self.assertEqual(self.header.Parameters.x, 10)

    def test_materials_block(self):
        """Test the structure of the Materials block"""
        self.assertTrue(hasattr(self.header, 'Parameters'))
        self.assertTrue(hasattr(self.header.Parameters, 'Materials'))
        self.assertCountEqual(self.header.Parameters.Materials.ids, [3])



# -*- coding: utf-8 -*-
"""
proc
====

The `simpleparse` library requires applications to define `dispatch processors` that are called for each token in the
grammar. This module defines the `dispatch processor`.

"""

import re

import numpy as np

# List of literals found in older files and the constans they should
# be mapped to ensuring consistency between different versions of
# AmiraMesh and HyperSurface files
_compatibilitymap = {
    'TetrahedronData': "Tetrahedra",
    'TetrahedraData': "Tetrahedra",
    'TriangleData': "Triangles",
    'NodeData': "Nodes",
    'EdgeData': "Edges",
    'LatticeData': "Lattice",
    'FieldData': "Field"
}

# removes leading and trailing whitespace and " characters from
# the strings defining string type parameters values such as names of
# materials
_strip_material_name = re.compile(r'^\s*"\s*|\s*",*$')

# simpleparse
from simpleparse.dispatchprocessor import DispatchProcessor, getString, dispatchList, singleMap


class AmiraDispatchProcessor(DispatchProcessor):
    """Class defining methods to handle each token specified in the grammar"""

    def designation(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        return {'designation': singleMap(value[3], self, buffer_)}

    def filetype(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def dimension(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def format(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def version(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def extra_format(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def comment(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        return {'comment': singleMap(value[3], self, buffer_)}

    def date(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def array_declarations(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        return {'array_declarations': dispatchList(self, value[3], buffer_)}

    def array_declaration(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        return singleMap(value[3], self, buffer_)

    def array_name(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def array_dimension(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        _av = dispatchList(self, value[3], buffer_)
        # if resulting list of sizes along all dimensions has only a single
        # entry than reduce it to a scalar size value otherwise convert to 
        # numpy.int64 array
        if len(_av) == 1:
            return int(_av[0])
        elif len(_av) > 1:
            return np.array(_av, dtype=np.int64)
        else:
            raise ValueError('definition value list is empty:', _av)

    def parameters(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        _av = singleMap(value[3], self, buffer_)
        if len(_av) == 1 and type(_av) in [dict] and "parameter_list" in _av:
            return {'parameters': _av["parameter_list"]}
        return {'parameters': _av}

    def materials(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        _av = dispatchList(self, value[3], buffer_)
        # currently gramma handles dedicated materials section as additional 
        # parameters section starting with Materials instead of Parameters.
        # rewrite it to materials struture to ensure they do not overwrite 
        # the preceeding parameters section
        return {
            'materials': [
                {
                    'parameter_name': _strip_material_name.sub('', _item[_id]['parameter_value']),
                    'parameter_value': _item[:_id] + _item[_id + 1:]
                }
                for _item, _id in (
                    (
                        _item,
                        [_index for _index, _val in enumerate(_item) if _val['parameter_name'] in ['name', 'Name']][0]
                    ) for _item in _av
                )
            ]
        }

    def parameter(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        return singleMap(value[3], self, buffer_)

    def parameter_value(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        _av = dispatchList(self, value[3], buffer_)
        if len(_av) == 1:
            return _av[0]
        return _av

    def parameter_list(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        return dispatchList(self, value[3], buffer_)

    def parameter_name(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def attribute_value(self, value, buffer_):  # @UnusedVariable
        # value = (tag, left, right, taglist)
        _av = dispatchList(self, value[3], buffer_)
        # if list of attribute values contains only one entry reduce it 
        # to scalar value. In case the list contains more than one value
        # insert the string "<!?c?!>" which will be used by the
        # AmiraHeader._load_parameters function to distinguisch between
        # a list of constant parameter values and a list of parameters
        # in case list of parameter values is empty return None
        if len(_av) == 1:
            return _av[0]
        elif len(_av) > 1:
            return ["<!?c?!>"] + _av
        elif len(_av) == 0:
            return None
        else:
            raise ValueError('attribute value list is empty:', _av)

    def inline_parameter_value(self, value, buffer_):
        # value = (tag, left, right, taglist)
        if value[3][0][0] == "qstring":
            return getString(value, buffer_).strip(' \t\n\r\f"')
        _av = dispatchList(self, value[3], buffer_)
        # if list of attribute values contains only one entry reduce it 
        # to scalar value. Otherwise # insert the string "<!?c?!>" which
        # will be used by the # AmiraHeader._load_parameters function to
        # distinguisch between a list of constant parameter values and a
        # list of sub parameters
        if len(_av) == 1:
            return _av[0]
        return ['<!?c?!>'] + _av

    def data_definitions(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return {'data_definitions': dispatchList(self, value[3], buffer_)}

    def data_definition(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return singleMap(value[3], self, buffer_)

    def interpolation_method(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def array_reference(self, value, buffer_):
        # value = (tag, left, right, taglist)
        reference = getString(value, buffer_)
        # rename older AmiraMesh array refrence specification to their
        # corresponding value found in newer files
        return _compatibilitymap.get(reference, reference)

    def data_type(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def data_dimension(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return int(getString(value, buffer_))

    def data_name(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def data_index(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return int(getString(value, buffer_))

    def data_format(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def data_length(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return int(getString(value, buffer_))

    def hyphname(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def xstring(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def qstring(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return getString(value, buffer_)

    def number(self, value, buffer_):
        # value = (tag, left, right, taglist)
        if value[3][0][0] == 'int':
            return int(getString(value, buffer_))
        elif value[3][0][0] == 'float':
            return float(getString(value, buffer_))
        else:
            return getString(value, buffer_)

    def number_seq(self, value, buffer_):
        # value = (tag, left, right, taglist)
        return dispatchList(self, value[3], buffer_)



# -*- coding: utf-8 -*-
"""
data_stream
===========

Classes that define data streams (DataList) in Amira (R) files

There are two main types of data streams:

* `AmiraMeshDataStream` is for `AmiraMesh` files
* `AmiraHxSurfaceDataStream` is for `HxSurface` files

Both classes inherit from `AmiraDataStream` class, which handles common functionality such as:

* initialisation with the header metadata
* the `get_data` method calls each subclass's `_decode` method

"""
from __future__ import print_function

import re
import sys
# todo: remove as soon as DataStreams class is removed
import warnings
import zlib

import numpy as np


from .core import _dict_iter_keys, _dict_iter_values, ListBlock, deprecated
from .grammar import _hyper_surface_file

# definition of numpy data types with dedicated endianess and number of bits
# they are used by the below lookup table
_np_ubytebig = np.dtype(np.uint8).newbyteorder('>')
_np_ubytelittle = np.dtype(np.uint8).newbyteorder('<')
_np_bytebig = np.dtype(np.int8).newbyteorder('>')
_np_bytelittle = np.dtype(np.int8).newbyteorder('<')
_np_shortlittle = np.dtype(np.int16).newbyteorder('<')
_np_shortbig = np.dtype(np.int16).newbyteorder('>')
_np_ushortlittle = np.dtype(np.uint16).newbyteorder('<')
_np_ushortbig = np.dtype(np.uint16).newbyteorder('>')
_np_intlittle = np.dtype(np.int32).newbyteorder('<')
_np_intbig = np.dtype(np.int32).newbyteorder('>')
_np_uintlittle = np.dtype(np.uint32).newbyteorder('<')
_np_uintbig = np.dtype(np.uint32).newbyteorder('>')
_np_longlittle = np.dtype(np.int64).newbyteorder('<')
_np_longbig = np.dtype(np.int64).newbyteorder('>')
_np_ulonglittle = np.dtype(np.uint64).newbyteorder('<')
_np_ulongbig = np.dtype(np.uint64).newbyteorder('>')
_np_floatbig = np.dtype(np.float32).newbyteorder('>')
_np_floatlittle = np.dtype(np.float32).newbyteorder('<')
_np_doublebig = np.dtype(np.float64).newbyteorder('>')
_np_doublelittle = np.dtype(np.float64).newbyteorder('<')
_np_complexbig = np.dtype(np.complex64).newbyteorder('>')
_np_complexlittle = np.dtype(np.complex64).newbyteorder('<')
_np_char = np.dtype(np.string_)

"""
lookuptable of the different data types occurring within
AmiraMesh and HyperSurface files grouped by their endianess
which is defined according to Amira Referenceguide [1] as follows

 * BINARY:               for bigendian encoded streams,

 * BINARY-LITTLE-ENDIAN: for little endian encoded streams and

 * ASCII:                for human readable encoded data

The lookup table is split into three sections:

 * True:  for all littleendian data types

 * False: for all bigendian data types and

 * the direct section mapping to the default numpy types for
   reading ASCII encoded data which have no specific endianness besides the
   bigendian characteristic intrinsic to decimal numbers.

[1] pp 519-525 # downloaded Dezember 2018 from 
   http://www1.udel.edu/ctcr/sites/udel.edu.ctcr/files/Amira%20Reference%20Guide.pdf
"""
_type_map = {
    True: {
        'byte': _np_ubytelittle,
        'ubyte': _np_ubytelittle,
        'short': _np_shortlittle,
        'ushort': _np_ushortlittle,
        'int': _np_intlittle,
        'uint': _np_uintlittle,
        'long': _np_longlittle,
        'ulong': _np_ulonglittle,
        'uint64': _np_ulonglittle,
        'float': _np_floatlittle,
        'double': _np_doublelittle,
        'complex': _np_complexlittle,
        'char': _np_char,
        'string': _np_char,
        'ascii': _np_char
    },
    False: {
        'byte': _np_ubytebig,
        'ubyte': _np_ubytebig,
        'short': _np_shortbig,
        'ushort': _np_ushortbig,
        'int': _np_intbig,
        'uint': _np_uintbig,
        'long': _np_longbig,
        'ulong': _np_ulongbig,
        'uint64': _np_ulongbig,
        'float': _np_floatbig,
        'double': _np_doublebig,
        'complex': _np_complexbig,
        'char': _np_char,
        'string': _np_char,
        'ascii': _np_char
    },
    'byte': np.dtype(np.int8),
    'ubyte': np.dtype(np.uint8),
    'short': np.dtype(np.int16),
    'ushort': np.dtype(np.uint16),
    'int': np.dtype(np.int32),
    'uint': np.dtype(np.uint32),
    'long': np.dtype(np.int64),
    'ulong': np.dtype(np.uint64),
    'uint64': np.dtype(np.uint64),
    'float': np.dtype(np.float32),
    'double': np.dtype(np.float64),
    'complex': np.dtype(np.complex64),
    'char': _np_char,
    'string': _np_char,
    'ascii': _np_char
}

# try to import native byterele_decoder binary and fallback to python implementation
try:
    # if import failed for whatever reason
    if sys.version_info[0] > 2:
        from ahds.decoders import byterle_decoder
    else:
        from .decoders import byterle_decoder
except ImportError:
    def byterle_decoder(data, output_size):
        """If the C-ext. failed to compile or is unimportable use this slower Python equivalent

        :param str data: a raw stream of data to be unpacked
        :param int output_size: the number of items when ``data`` is uncompressed
        :return np.array output: an array of ``np.uint8``
        """

        from warnings import warn
        warn("using pure-Python (instead of Python C-extension) implementation of byterle_decoder")

        input_data = np.frombuffer(data, dtype=_np_ubytelittle, count=len(data))
        output = np.zeros(output_size, dtype=np.uint8)
        i = 0
        count = True
        repeat = False
        no = None
        j = 0
        while i < len(input_data):
            if count:
                no = input_data[i]
                if no > 127:
                    no &= 0x7f  # 2's complement
                    count = False
                    repeat = True
                    i += 1
                    continue
                else:
                    i += 1
                    count = False
                    repeat = False
                    continue
            elif not count:
                if repeat:
                    # value = input_data[i:i + no]
                    repeat = False
                    count = True
                    # output[j:j+no] = np.array(value)
                    output[j:j + no] = input_data[i:i + no]
                    i += no
                    j += no
                    continue
                elif not repeat:
                    # value = input_data[i]
                    # output[j:j+no] = value
                    output[j:j + no] = input_data[i]
                    i += 1
                    j += no
                    count = True
                    repeat = False
                    continue

        assert j == output_size

        return output

# define common alias for the selected byterle_decoder implementation
hxbyterle_decode = byterle_decoder


def hxzip_decode(data, output_size):
    """Decode HxZip data stream

    :param str data: a raw stream of data to be unpacked
    :param int output_size: the number of items when ``data`` is uncompressed
    :return np.array output: an array of ``np.uint8``
    """
    return np.frombuffer(zlib.decompress(data), dtype=_np_ubytelittle, count=output_size)


def set_data_stream(name, header):
    """Factory function used by AmiraHeader to determine the type of data stream present"""
    if header.filetype == 'AmiraMesh':
        return AmiraMeshDataStream(name, header)
    elif header.filetype == 'HyperSurface':
        return AmiraHxSurfaceDataStream(name, header)


class AmiraDataStream(ListBlock):
    """"""
    __slots__ = ('_stream_data', '_header')

    def __init__(self, name, header):
        self._header = header  # contains metadata for extracting streams
        self._stream_data = None
        super(AmiraDataStream, self).__init__(name)

    @property
    def load_stream(self):
        """Reports whether data streams are loaded or not"""
        return self._header.load_streams

    def get_data(self):
        """Decode and return the stream data in this stream"""
        try:
            assert len(self._stream_data) > 0
        except AssertionError:
            raise ValueError('empty stream found')
        return self._decode(self._stream_data)


class AmiraMeshDataStream(AmiraDataStream):
    """Class that defines an AmiraMesh data stream"""

    def read(self):
        """Extract the data streams from the AmiraMesh file"""
        with open(self._header.filename, 'rb') as f:
            # rewind the file pointer to the end of the header
            f.seek(len(self._header))
            start = int(self.data_index)  # this data streams index
            end = start + 1
            if self._header.data_stream_count == start:  # this is the last stream
                data = f.read()
                _regex = ".*?\n@{}\n(?P<stream>.*)\n".format(start).encode('ASCII')
                regex = re.compile(_regex, re.S)
                match = regex.match(data)
                _stream_data = match.group('stream')
                self._stream_data = _stream_data[:-1] if _stream_data[-1] == 10 else _stream_data
            else:
                data = f.read()
                _regex = ".*?\n@{}\n(?P<stream>.*)\n@{}\n".format(start, end).encode('ASCII')
                regex = re.compile(_regex, re.S)
                match = regex.match(data)
                self._stream_data = match.group('stream')

    def _decode(self, data):
        """Performs data stream decoding by introspecting the header information"""
        # determine the new output shape
        # take into account shape and dimension
        if isinstance(self.shape, tuple):
            if self.dimension > 1:
                new_shape = tuple(list(self.shape) + [self.dimension])
            else:
                new_shape = (self.shape, )
        elif isinstance(self.shape, int):
            if self.dimension > 1:
                new_shape = tuple([self.shape, self.dimension])
            else:
                new_shape = (self.shape, )
        # first we handle binary files
        # NOTE ON HOW LATTICES ARE STORED
        # AmiraMesh files state the dimensions of the lattice as nx, ny, nz
        # The data is stored such that the first value has index (0, 0, 0) then
        # (1, 0, 0), ..., (nx-1, 0, 0), (0, 1, 0), (1, 1, 0), ..., (nx-1, 1, 0)
        # In other words, the first index changes fastest.
        # The last index is thus the stack index
        # (see pg. 720 of https://assets.thermofisher.com/TFS-Assets/MSD/Product-Guides/user-guide-amira-software.pdf
        if self._header.format == 'BINARY':
            # _type_map[endianness] uses endianness = True for endian == 'LITTLE'
            is_little_endian = self._header.endian == 'LITTLE'
            if self.format is None:
                return np.frombuffer(
                    data,
                    dtype=_type_map[is_little_endian][self.type]
                ).reshape(*new_shape)
            elif self.format == 'HxZip':
                return np.frombuffer(
                    zlib.decompress(data),
                    dtype=_type_map[is_little_endian][self.type]
                ).reshape(*new_shape)
            elif self.format == 'HxByteRLE':
                size = int(np.prod(np.array(self.shape)))
                return hxbyterle_decode(
                    data,
                    size
                ).reshape(*new_shape)
            else:
                raise ValueError('unknown data stream format: \'{}\''.format(self.format))
        # explicit instead of assumption
        elif self._header.format == 'ASCII':
            return np.fromstring(
                data,
                dtype=_type_map[self.type],
                sep="\n \t"
            ).reshape(*new_shape)
        else:
            raise ValueError("unknown file format: {}".format(self._header.format))


class AmiraHxSurfaceDataStream(AmiraDataStream):
    """Class that defines an Amira HxSurface data stream"""

    def read(self):
        """Extract the data streams from the HxSurface file"""
        with open(self._header.filename, 'rb') as f:
            # rewind the file pointer to the end of the header
            f.seek(len(self._header))
            data = f.read()
            # get the vertex count and streams
            _vertices_regex = r".*?\n" \
                              r"Vertices (?P<vertex_count>\d+)\n" \
                              r"(?P<streams>.*)".encode('ASCII')
            vertices_regex = re.compile(_vertices_regex, re.S)
            match_vertices = vertices_regex.match(data)
            # todo: fix for full.surf and simple.surf
            # print(f"streams: {match_vertices.group('streams')}")
            vertex_count = int(match_vertices.group('vertex_count'))
            # get the patches
            # fixme: general case for NBranchingPoints, NVerticesOnCurves, BoundaryCurves being non-zero
            stream_regex = r"(?P<vertices>.*?)\n" \
                           r"NBranchingPoints (?P<branching_point_count>\d+)\n" \
                           r"NVerticesOnCurves (?P<vertices_on_curves_count>\d+)\n" \
                           r"BoundaryCurves (?P<boundary_curve_count>\d+)\n" \
                           r"Patches (?P<patch_count>\d+)\n" \
                           r"(?P<patches>.*)".encode('ASCII')
            match_streams = re.match(stream_regex, match_vertices.group('streams'), re.S)
            # instatiate the vertex block
            vertices_block = AmiraHxSurfaceDataStream('Vertices', self._header)
            # set the data for this stream
            vertices_block._stream_data = match_streams.group('vertices')
            # length, type and dimension are needed for decoding
            vertices_block.add_attr('length', vertex_count)
            vertices_block.add_attr('type', 'float')
            vertices_block.add_attr('dimension', 3)
            vertices_block.add_attr('data', vertices_block.get_data())
            vertices_block.add_attr('NBranchingPoints', 0)
            vertices_block.add_attr('NVerticesOnCurves', 0)
            vertices_block.add_attr('BoundaryCurves', 0)
            # instantiate the patches block
            patches_block = AmiraHxSurfaceDataStream('Patches', self._header)
            patch_count = int(match_streams.group('patch_count'))
            patches_block.add_attr('length', patch_count)
            # get the triangles and contents of each patch
            # fixme: general case for BoundaryID, BranchingPoints being non-zero
            #  i've not seen an example with loaded fields
            # todo: consider compiling regular expressions
            # NOTE:
            # There is a subtlety with this regex:
            # It might be the case that the last part matches bytes that end in '\n[}]\n'
            # that are not the end of the stream. The only way to remede this is to include the
            # extra brace [{] so that it now matches '\n[}]\n[{]', which is more likely to
            # correspond to the end of the patch. However this introduces a problem:
            # we will not be able to match the last patch unless we also add [{] to the stream to match.
            # This also means that start_from argument will be wrong given that it will have past
            # the starting point of the next patch. This is trivial to solve because we simply
            # backtrack start_from by 1.
            # These are noted in NOTE A and NOTE B below.
            _patch_regex = r"[{]\n" \
                           r"InnerRegion (?P<patch_inner_region>.*?)\n" \
                           r"OuterRegion (?P<patch_outer_region>.*?)\n" \
                           r"BoundaryID (?P<patch_boundary_id>\d+)\n" \
                           r"BranchingPoints (?P<patch_branching_points>\d+)\n" \
                           r"\s+\n" \
                           r"Triangles (?P<triangle_count>.*?)\n" \
                           r"(?P<triangles>.*?)\n" \
                           r"[}]\n[{]".encode('ASCII')
            patch_regex = re.compile(_patch_regex, re.S)
            # start from the beginning
            start_from = 0
            for p_id in range(patch_count):
                # NOTE A
                match_patch = patch_regex.match(match_streams.group('patches') + b'{', start_from)
                patch_block = AmiraHxSurfaceDataStream('Patch', self._header)
                patch_block.add_attr('InnerRegion', match_patch.group('patch_inner_region').decode('utf-8'))
                patch_block.add_attr('OuterRegion', match_patch.group('patch_outer_region').decode('utf-8'))
                patch_block.add_attr('BoundaryID', int(match_patch.group('patch_boundary_id')))
                patch_block.add_attr('BranchingPoints', int(match_patch.group('patch_branching_points')))
                # let's now add the triangles from the patch
                triangles_block = AmiraHxSurfaceDataStream('Triangles', self._header)
                # set the raw data stream
                triangles_block._stream_data = match_patch.group('triangles')
                # decoding needs to have the length, type, and dimension
                triangles_block.add_attr('length', int(match_patch.group('triangle_count')))
                triangles_block.add_attr('type', 'int')
                triangles_block.add_attr('dimension', 3)
                # print('debug:', int(match_patch.group('triangle_count')), len(match_patch.group('triangles')))
                # print('debug:', match_patch.group('triangles')[:20])
                # print('debug:', match_patch.group('triangles')[-20:])
                triangles_block.add_attr('data', triangles_block.get_data())
                # now we can add the triangles block to the patch...
                patch_block.add_attr(triangles_block)
                # then we collate the patches
                patches_block.append(patch_block)
                # the next patch begins where the last patch ended
                # NOTE B
                start_from = match_patch.end() - 1  # backtrack by 1
            # add the patches to the vertices
            vertices_block.add_attr(patches_block)
            # add the vertices to the data stream
            self.add_attr(vertices_block)

    def _decode(self, data):
        is_little_endian = self._header.endian == 'LITTLE'
        if self._header.format == 'BINARY':
            return np.frombuffer(
                data,
                dtype=_type_map[is_little_endian][self.type]
            ).reshape(self.length, self.dimension)
        elif self._header.format == 'ASCII':
            return np.fromstring(
                data,
                dtype=_type_map[self.type],
                sep="\n \t"
            ).reshape(self.length, self.dimension)


@deprecated(
    "DataStreams class is obsolete, access data using stream_data and data attributes of corresponding metadata block attributes of AmiraHeader instance")
class DataStreams(object):
    __slots__ = ("_header", "__stream_data")

    def __init__(self, header):
        self._header = header
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            if self._header.filetype == "AmiraMesh":
                self.__stream_data = self._header.data_pointers
            else:
                self.__stream_data = dict()
                for _streamblock in _dict_iter_keys(_hyper_surface_file):
                    _streamlist = _streamblock  # self._header._stream_loader.__class__._group_array_map.get(_streamblock,_streamblock).format('List')
                    _streamlist = getattr(self._header, _streamlist, None)
                    if _streamlist is None:
                        continue
                    self.__stream_data[_streamblock] = _streamlist

    def __getattribute__(self, attr):
        if attr in ("file", "header", "stream_data", "filetype"):
            return super(DataStreams, self).__getattribute__(attr)()
        return super(DataStreams, self).__getattribute__(attr)

    @deprecated("use <AmiraHeader>.filename instead")
    def file(self):
        return self._header.filename

    @deprecated("use AmiraHeader instance directly")
    def header(self):
        return self._header

    @deprecated(
        "access data of individual streams through corresponding attributes and dedicated stream_data and data attributes of meta data blocks")
    def stream_data(self):
        return self.__stream_data

    @deprecated("use <AmiraHeader>.filetype attribute instead")
    def filetype(self):
        return self._header.filetype

    @deprecated
    def __len__(self):
        return len(self.__stream_data)

    @deprecated
    def __iter__(self):
        return iter(_dict_iter_values(self.__stream_data))

    @deprecated
    def __getitem__(self, key):
        return self.__stream_data[key]

    @deprecated
    def __repr__(self):
        return "{} object with {} stream(s): {} ".format(
            self.__class__,
            len(self),
            ", ".join(_dict_iter_keys(self.__stream_data))
        )



# -*- coding: utf-8 -*-
"""
core
====

Core functionality for `ahds` package

Contains:

* Python2/3 adapters
* decorator for marking deprecation
* core base classes: `Block` and `ListBlock`

"""
from __future__ import print_function

import functools as ft
import inspect
import sys
import warnings

import numpy as np

# print to stderr
_print = ft.partial(print, file=sys.stderr)

if sys.version_info[0] > 2:
    # All definitions for Python3 and newer which differ from their counterparts in Python2.x
    def _decode_string(data):
        """ decodes binary ASCII string to python3 UTF-8 standard string """
        try:
            return data.decode("ASCII")
        except:
            return data.decode("UTF8")


    # in define xrange alias which has been removed in Python3 as range now is equal to
    xrange = range  # xrange
    # define _dict_iter_values, _dict_iter_items, _dict_iter_keys aliases imported by the
    # other parts for dict.values, dict.items and dict.keys
    _dict_iter_values = dict.values
    _dict_iter_items = dict.items
    _dict_iter_keys = dict.keys
    if sys.version_info[1] >= 7:
        _dict = dict
    else:
        from collections import OrderedDict

        _dict = OrderedDict


    def _qualname(o):
        return o.__qualname__


    _str = str

    # UserList
    from collections import UserList
    _UserList = UserList

else:
    import __builtin__


    # Python2.x definitions
    def _decode_string(data):
        """No decoding in Python2.x"""
        return data


    # try to define xrange alias pointing to the builtin xrange
    try:
        xrange = __builtins__['xrange']
    except AttributeError:
        xrange = getattr(__builtins__, 'xrange')
    # define _dict_iter_values, _dict_iter_items, _dict_iter_keys aliases imported by the
    # other parts for dict.itervalues, dict.iteritems and dict.iterkeys
    _dict_iter_values = dict.itervalues
    _dict_iter_items = dict.iteritems
    _dict_iter_keys = dict.iterkeys
    from collections import OrderedDict

    _dict = OrderedDict

    def _qualname(o):
        return o.__name__

    _str = __builtin__.unicode

    # UserList
    from UserList import UserList
    _UserList = UserList


def deprecated(description):
    """Function/Class/Method decorator for warning about deprecations"""
    def outer_wrapper(o):
        @ft.wraps(o)
        def inner_wrapper(*args, **kwargs):
            if inspect.isfunction(o):
                warnings.warn('function/method {} is deprecated: {}'.format(_qualname(o), description),
                              DeprecationWarning)
            elif inspect.isclass(o):
                warnings.warn('class {} is deprecated: {}'.format(_qualname(o), description), DeprecationWarning)
            return o(*args, **kwargs)

        return inner_wrapper

    return outer_wrapper


@ft.total_ordering
class Block(object):
    """Data content block for atomic entities"""
    __slots__ = ('_name', '_attrs', '_is_parent', '__dict__', '__weakref__')

    def __init__(self, name):
        self._name = name
        self._attrs = _dict()
        self._is_parent = False

    @property
    def name(self):
        return self._name

    def attrs(self):
        return list(self._attrs.keys())

    @property
    def is_parent(self):
        return self._is_parent

    def add_attr(self, attr, value=None, isparent=False):
        """Add an attribute to this block object"""
        try:
            assert hasattr(attr, 'name') or isinstance(attr, str)
        except AssertionError:
            raise ValueError('attr should be str or have .name attribute')
        if hasattr(attr, 'name'):
            value = attr
            attr = attr.name
        elif not isinstance(attr, str):
            raise ValueError("invalid type for attr: {}".format(type(attr)))
        # first check that the attribute does not exist on the class
        if hasattr(self, attr):
            raise ValueError("will not overwrite attribute '{}'".format(attr))
        try:
            assert attr not in self._attrs
        except AssertionError:
            raise ValueError("attribute '{}' already exists".format(attr))

        if isinstance(value, Block):
            self._attrs[attr] = value
            self._is_parent = True
        else:
            self._attrs[attr] = value

    def __setattr__(self, key, value):
        """Guard against unintentional modification of _attrs"""
        if key == '_attrs':  # we can't prevent it but we can control it's type and content
            # value must be a dictionary
            try:
                assert isinstance(value, dict)
            except AssertionError:
                raise ValueError('{} must be a dict'.format(key))
            # value must have strings for keys
            try:
                keys = list(value.keys())
                keys_are_strings = map(lambda x: isinstance(x, str), keys)
                assert all(keys_are_strings)  # or len(keys) == 0
            except AssertionError:
                raise ValueError("all keys of {} must be strings".format(key))
        super(Block, self).__setattr__(key, value)

    # todo: rename to 'rename_attr'
    # todo: change signature to rename(self, name, new_name)
    def move_attr(self, new_name, name):
        """Rename an attribute"""
        try:
            assert new_name not in self._attrs
        except AssertionError:
            raise ValueError("will not overwrite attribute '{}'".format(new_name))
        else:
            try:
                self._attrs[new_name] = self._attrs[name]
                del self._attrs[name]
            except KeyError:
                raise AttributeError('''no attribute '{}' found'''.format(name))

    def rename_attr(self, attr, new_name):
        self.move_attr(new_name, attr)

    def __getattr__(self, name):
        try:
            return self._attrs[name]
        except KeyError:
            raise AttributeError('''attribute {} not found'''.format(name))

    def __str__(self, prefix="", index=None,alt_name=None):
        """Compile the hierarchy of Blocks into a tree

        :param str prefix: prefix to signify depth in the tree
        :param int index: applies for list items [default: None]
        :returns str string: formatted string of attributes
        """
        # the root Block will have an empty prefix
        # but the prefix will be updated calls for __str__ for nested Blocks
        # we use the format() function to pass a format_spec which does alignment
        string = ''
        if index is not None:
            string += "{} {} [is_parent? {:<5}]\n".format(
                format(prefix + "+[{}]-{}".format(index, self.name if alt_name in (self.name,None,'') else alt_name), '<55'),
                format(type(self).__name__, '>50'),
                str(self.is_parent)
            )
        else:
            name = format(prefix + "+-{}".format(self.name if alt_name in (self.name,None,'') else alt_name), '<55')
            if len(name) > 55:
                name = name[:52] + '...'
            string += "{} {} [is_parent? {:<5}]\n".format(
                name,
                format(type(self).__name__, '>50'),
                str(self.is_parent)
            )
        for attr in self._attrs:
            # check if the attribute is Block or non-Block
            if isinstance(self._attrs[attr], Block):
                # if it is a Block then the prefix will change by having extra '| ' before it
                # string += 'something\n'
                string += self._attrs[attr].__str__(prefix=prefix + "|  ",alt_name=attr)
            else:
                # if it is not a Block then we construct the repr. manually
                val = self._attrs[attr]
                # don't print the whole array for large arrays
                if isinstance(val, (np.ndarray,)):
                    # we construct a tuple for the first array element (0,...,0) and the last
                    # one (-1,...,-1); however, we have to do this independent of the dimensions
                    # we use a tuple constructed using shape - 1 in both cases
                    start = tuple([0] * (len(val.shape) - 1))
                    end = tuple([-1] * (len(val.shape) - 1))
                    if start == end:
                        string += prefix + "|  +-{}: {}\n".format(attr, val[start])
                    else:
                        string += prefix + "|  +-{}: {},...,{}\n".format(attr, val[start], val[end])
                else:
                    if isinstance(self._attrs[attr], str):
                        if attr == "@alias" and alt_name == self._attrs[attr]:
                            string += prefix + "|  +-{}: {}\n".format(attr, self.name)
                        elif len(self._attrs[attr]) > 55:
                            string += prefix + "|  +-{}: {}\n".format(attr, self._attrs[attr][:52] + '...')
                        else:
                            string += prefix + "|  +-{}: {}\n".format(attr, self._attrs[attr])
                    else:
                        string += prefix + "|  +-{}: {}\n".format(attr, self._attrs[attr])
        return string

    def __getitem__(self, index):
        try:
            assert isinstance(index, int)
        except AssertionError:
            raise ValueError('index must be an integer or long')
        if self.name == 'Materials':
            for attr in self.attrs():
                block = getattr(self, attr)
                if hasattr(block, 'Id'):
                    if getattr(block, 'Id') == index:
                        return block
                    else:
                        continue
        return

    def __contains__(self, item):
        if item in self._attrs:
            return True
        return False

    def __eq__(self, other):
        try:
            assert isinstance(other, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        if self.name == other.name:
            return True
        return False

    def __le__(self, other):
        try:
            assert isinstance(other, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        if self.name < other.name:
            return True
        return False


class ListBlock(Block):
    """Data content block for sequence entities"""
    __slots__ = ('_list', '_material_dict')

    def __init__(self, *args, **kwargs):
        super(ListBlock, self).__init__(*args, **kwargs)
        self._list = list()  # separate attribute for ease of management
        self._material_dict = dict()  # a dictionary used by Materials to extract material by material name

    def items(self):
        return self._list

    @property
    def ids(self):
        ids = list()
        if self.name == 'Materials':
            # first check non-list items
            for attr in self.attrs():
                _attrval = getattr(self, attr)
                if hasattr(_attrval, 'Id'):
                    ids.append(int(_attrval.Id))
                elif hasattr(_attrval, 'id'):
                    ids.append(int(_attrval.id))
            # then check list items
            for list_item in self:
                if hasattr(list_item, 'Id'):
                    ids.append(int(list_item.Id))
                elif hasattr(list_item, 'id'):
                    ids.append(int(list_item.id))
        return ids

    @property
    def material_dict(self):
        """A convenience dictionary of materials indexed by material name

        If this is not a Materials ListBlock (name = 'Material') then it should return None
        """
        # todo: testcase: name == 'Materials' ? dictionary of material blocks : None
        return self._material_dict

    @material_dict.setter
    def material_dict(self, value):
        """Check that this is a material block"""
        if self.name == "Materials":
            if isinstance(value, dict):
                keys_are_strings = map(lambda k: isinstance(k, str), value.keys())
                values_are_blocks = map(lambda v: isinstance(v, Block), value.values())
                try:
                    assert all(keys_are_strings)
                except AssertionError:
                    raise ValueError("keys for material_dict dictionary must be strings")
                try:
                    assert all(values_are_blocks)
                except AssertionError:
                    raise ValueError("values for material_dict dictionary must be Blocks (or subclasses)")
                # now we can set
                self._material_dict = value
            else:
                raise TypeError("value must be a dict")
        else:
            raise ValueError("the material_dict attribute can only be set for Materials ListBlocks")

    def __setattr__(self, key, value):
        """We do some sanity checks before allowing direct setting"""
        # we only allow modification of _list if it meets some criteria
        if key == '_list':
            # it must be a list
            try:
                assert isinstance(value, list)
            except AssertionError:
                raise ValueError("_list attribute must be a list")
            # make sure it's either empty or has block subclasses
            if len(value) > 0:
                try:
                    assert all(map(lambda x: isinstance(x, Block), value))
                except AssertionError:
                    raise ValueError("list contains non-Block class/subclass")
        super(ListBlock, self).__setattr__(key, value)

    @property
    def is_parent(self):
        """A ListBlock is a parent if it has a Block attribute or if it has list items"""
        # todo: testcase
        if super(ListBlock, self)._is_parent:
            return True
        else:
            if len(self._list) > 0:
                return True
            else:
                return False

    def __str__(self, prefix="", index=None,alt_name=None):
        """Convert the ListBlock into a string

        :param str prefix: prefix to signify depth in the tree
        :param int index: applies for list items [default: None]
        :returns str string: formatted string of attributes
        """
        # first we use the superclass to populate everything else
        string = super(ListBlock, self).__str__(prefix=prefix, index=index,alt_name=alt_name)
        # now we stringify the list-blocks
        for index, block in enumerate(self.items()):
            string += block.__str__(prefix=prefix + "|  ", index=index)
        return string

    def __len__(self):
        return len(self._list)

    def __setitem__(self, key, value):
        try:
            assert isinstance(value, Block)
        except AssertionError:
            raise ValueError('value must be a Block class/subclass')
        try:
            self._list[key] = value
        except IndexError:
            self._list.append(value)
            # raise ValueError("index {} does not exist".format(key))

    def __getitem__(self, item):
        try:
            return self._list[item]
        except KeyError:
            raise IndexError("no item with index '{}'".format(item))

    def __iter__(self):
        return iter(self._list)

    def __contains__(self, item):
        if item in self._list:
            return True
        return False

    def __delitem__(self, key):
        try:
            del self._list[key]
        except KeyError:
            raise IndexError("missing item at index'{}'".format(key))

    # Mutable sequences should provide methods append(), count(), index(), extend(), insert(), pop(), remove(),
    # reverse() and sort(), like Python standard list objects.
    def append(self, item):
        try:
            assert isinstance(item, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        self._list.append(item)

    def count(self, item, *args):
        try:
            assert isinstance(item, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        return self._list.count(item, *args)

    def index(self, item, *args):
        try:
            assert isinstance(item, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        return self._list.index(item, *args)

    def extend(self, item):
        try:
            assert isinstance(item, list)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        self._list.extend(item)

    def insert(self, index, item):
        try:
            assert isinstance(item, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        self._list.insert(index, item)

    def pop(self, *args):
        return self._list.pop(*args)

    def remove(self, item):
        try:
            assert isinstance(item, Block)
        except AssertionError:
            raise ValueError('item must be a Block class/subclass')
        return self._list.remove(item)

    def reverse(self):
        self._list.reverse()

    def sort(self, **kwargs):
        return self._list.sort(**kwargs)



# -*- coding: utf-8 -*-
"""
extra
=====

Additional functionality that is not currently used and may be deprecated in future.

* process an AmiraMesh lattice as a stack of images (`Image` and `ImageSet`)
* compute contours around segment for each image (`Contour` and `ContourSet`)

"""

import numpy as np
from skimage.measure._find_contours import find_contours

from .core import (
    _dict_iter_items, _dict_iter_values, xrange, _UserList
)


class Image(object):
    """Encapsulates individual images"""

    def __init__(self, z, array):
        self.z = z
        self._array = array
        self._byte_values = np.unique(self._array)

    def __getattribute__(self, attr):
        if attr in ("array", "byte_values"):
            return super(Image, self).__getattribute__("_" + attr)
        if attr in ("as_contours", "as_segments"):
            return super(Image, self).__getattribute__("_" + attr)()
        return super(Image, self).__getattribute__(attr)

    def equalise(self):
        """Increase the dynamic range of the image"""
        multiplier = 255 // len(self._byte_values)
        return self._array * multiplier

    def _as_contours(self):
        """A dictionary of lists of contours keyed by byte_value"""
        contours = dict()
        _maskbase = np.array([False, True])
        _indexbase = np.zeros(self._array.shape, dtype=np.int8)
        for byte_value in self._byte_values[self._byte_values != 0]:
            mask = _maskbase[np.equal(self._array, byte_value, out=_indexbase)]
            found_contours = find_contours(mask, 254, fully_connected='high')  # a list of array
            contours[byte_value] = ContourSet(found_contours)
        return contours

    def _as_segments(self):
        return {self.z: self.as_contours}

    def show(self):
        """Display the image"""
        with_matplotlib = True
        try:
            import matplotlib.pyplot as plt
        except RuntimeError:
            import skimage.io as io
            with_matplotlib = False

        if with_matplotlib:
            equalised_img = self.equalise()

            _, ax = plt.subplots()

            ax.imshow(equalised_img, cmap='gray')

            import random

            for contour_set in _dict_iter_values(self.as_contours):
                r, g, b = random.random(), random.random(), random.random()
                [ax.plot(contour[:, 1], contour[:, 0], linewidth=2, color=(r, g, b, 1)) for contour in contour_set]

            ax.axis('image')
            ax.set_xticks([])
            ax.set_yticks([])

            plt.show()
        else:
            io.imshow(self.equalise())
            io.show()

    def __repr__(self):
        return "<Image with dimensions {}>".format(self.array.shape)

    def __str__(self):
        return "<Image with dimensions {}>".format(self.array.shape)


class ImageSet(_UserList):
    """Encapsulation for set of ``Image`` objects"""

    def __getitem__(self, index):
        return Image(index, self.data[index])

    def __getattribute__(self, attr):
        if attr in ("segments",):
            return super(ImageSet, self).__getattribute__("_" + attr)()
        return super(ImageSet, self).__getattribute__(attr)

    def _segments(self):
        """A dictionary of lists of contours keyed by z-index"""
        segments = dict()
        for i in xrange(len(self)):
            image = self[i]
            for z, contour in _dict_iter_items(image.as_segments):
                for byte_value, contour_set in _dict_iter_items(contour):
                    if byte_value not in segments:
                        segments[byte_value] = dict()
                    if z not in segments[byte_value]:
                        segments[byte_value][z] = contour_set
                    else:
                        segments[byte_value][z] += contour_set

        return segments

    def __repr__(self):
        return "<ImageSet with {} images>".format(len(self))


class ContourSet(_UserList):
    """Encapsulation for a set of ``Contour`` objects"""

    def __getitem__(self, index):
        return Contour(index, self.data[index])

    def __repr__(self):
        string = "{} with {} contours".format(self.__class__, len(self))
        return string


class Contour(object):
    """Encapsulates the array representing a contour"""

    def __init__(self, z, array):
        self.z = z
        self._array = array

    def __len__(self):
        return len(self._array)

    def __iter__(self):
        return iter(self._array)

    @staticmethod
    def string_repr(self):
        string = "<Contour at z={} with {} points>".format(self.z, len(self))
        return string

    def __repr__(self):
        return self.string_repr(self)

    def __str__(self):
        return self.string_repr(self)



# -*- coding: utf-8 -*-
# amira_grammar_parser.py
"""
grammar
=======

We define an EBNF grammar for Amira (R) headers to extract all metadata. In addition to that,
we also define how `HxSurface` files are structured.

This module also includes several helper functions that use the grammar resources:

*   the `get_header` function returns only the header up to the first data stream; data is returned as a decoded string (`UTF-8`);

*   the `parse_header` function applies the grammar to return a nested set of Python primitives to be transformed into an `AmiraHeader` object;

*   the `get_parsed_data` function transparently applied both above functions given the Amira (R) filename

"""

from __future__ import print_function

import re
import sys

# simpleparse
from simpleparse.parser import Parser
from simpleparse.common import numbers, strings
from simpleparse.dispatchprocessor import DispatchProcessor, getString, dispatchList, dispatch, singleMap, multiMap

from .core import _decode_string, _dict_iter_items, _dict_iter_keys
from .proc import AmiraDispatchProcessor

# on autoformat these two lines disappear; adding them here in case that happens
# from simpleparse.common import numbers, strings
# from simpleparse.dispatchprocessor import DispatchProcessor, getString, dispatchList, dispatch, singleMap, multiMap


# Amira (R) Header Grammar
amira_header_grammar = (r'''
amira                        :=    designation, tsn, comment*, tsn*, array_declarations, tsn, parameters*, materials*, data_definitions, tsn

designation                  :=    ("#", ts, filetype, ts, dimension*, ts*, format, ts, version, ts*, extra_format*, tsn) / ("#", ts, filetype, ts, version, ts, format, tsn)
filetype                     :=    "AmiraMesh" / "HyperSurface" / "Avizo"
dimension                    :=    "3D"
format                       :=    "BINARY-LITTLE-ENDIAN" / "BINARY" / "ASCII"
version                      :=    number
extra_format                 :=    "<", "hxsurface", ">"

comment                      :=    ts, ("#", ts, "CreationDate:", ts, date) / ("#", ts, xstring) , tsn
date                         :=    xstring

array_declarations           :=    array_declaration*
array_declaration            :=    ("define", ts, array_name, ts, array_dimension) / ("n", array_name, ts, array_dimension), tsn
array_name                   :=    hyphname
array_dimension              :=    number, (ts, number)*

parameters                   :=    "Parameters" , ts, parameter_list, tsn 
parameter                    :=    ts, parameter_name, ts, parameter_value, c*, tsn
parameter_name               :=    hyphname
parameter_value              :=    parameter_list / inline_parameter_value / attribute_value
parameter_list               :=    "{", tsn, ( parameter / comment )*, "}" 
attribute_value              :=    ("-"*, "\""*, ((number, (ts, number)*) / xstring)*, "\""*)
inline_parameter_value       :=    (number, (ts, number)*) / qstring

materials                    :=    "Materials" , tsn, "{", tsn*, ( parameter_list, tsn* )+, "}", tsn 

data_definitions             :=    data_definition*
data_definition              :=    array_reference , ts, "{", ts, data_type, "["*, data_dimension*, "]"*, ts, data_name, ts, "}", ts, "="*, ts*, interpolation_method*, "("*, "@", data_index, ")"* , "("*, data_format*, ","*, data_length* , ")"* , tsn
array_reference              :=    hyphname / "Field"
data_type                    :=    hyphname
data_dimension               :=    number
data_name                    :=    hyphname
data_index                   :=    number
data_format                  :=    "HxByteRLE" / "HxZip"
data_length                  :=    number
interpolation_method         :=    "Linear" / "Constant" / "EdgeElem"

hyphname                     :=    [A-Za-z_&], [-A-Za-z0-9_:]*
qstring                      :=    "\"", "["*, [-A-Za-z0-9_,.:/ tÅ�$;\n]*, "]"*, "\""
xstring                      :=    [A-Za-z], [A-Za-z0-9_\- �(\xc5)]*
number_seq                   :=    number, (ts, number)*

# silent production rules
<tsn>                        :=    [ \t\n]*            
<ts>                         :=    [ \t]*
<c>                          :=    ","
'''
                        )

# dict representing structure of hypersurface file according to Amira Reference guide
# pp 519-525 # downloaded Dezember 2018 from 
# http://www1.udel.edu/ctcr/sites/udel.edu.ctcr/files/Amira%20Reference%20Guide.pdf
_hyper_surface_file = {
    'Vertices': ['Coordinates', 3, 'float', False],
    'NBranchingPoints': [None, None, 'int', True],
    'NVerticesOnCurves': [None, None, 'int', True],
    'BoundaryCurves': {
        'Vertices': [None, 1, 'int', False],
        0: True
    },
    'Patches': {
        'InnerRegion': [None, None, 'str', False],
        'OuterRegion': [None, None, 'str', False],
        'Triangles': [None, 3, 'int', False],
        'BranchingPoints': [None, 0, 'int', True],
        'BoundaryCurves': [None, 0, 'int', True],
        0: False
    },
    'Surfaces': {
        'Region': [None, None, 'str', False],
        'Patches': [None, 0, 'int', False],
        0: True
    }
}

# string representing all valid keys within the above structure is inserted 
# in the below regular expression patterns
# todo: replace this with something more meaningful
_hyper_surface_entities = '|'.join(['|'.join(
    [_key] + ([_vk for _vk in _dict_iter_keys(_val) if isinstance(_vk, str)] if isinstance(_val, dict) else [])) for
    _key, _val in _dict_iter_items(_hyper_surface_file) if isinstance(_key, str)])

# maximum number of bytes to be rescanned at the end of the already inspected
# _stream_data array after new bytes have been read from the file. In case within this
# range a data stream marker (@<Num>) or any of the above HyperSurface section keys has
# alreday been successfully identified rescan starts at the byte following this match
# todo: replace this with something more meaningful
_rescan_overlap = max((
    max([len(_key)] + (
        [len(_vk) for _vk in _dict_iter_keys(_val) if isinstance(_vk, str)] if isinstance(_val, dict) else []))
    for _key, _val in _dict_iter_items(_hyper_surface_file)
    if isinstance(_key, str)
)) + 16

if sys.version_info[0] > 2:
    # definitions required by python3 and newer for properly parsing binary byte strings without converting
    # them first into regular UTF-8 strings

    _file_format_match = (
        re.compile(b'.*AmiraMesh.*'),
        re.compile(b'.*HyperSurface.*'),
        re.compile(b'.*Avizo.*')
    )

    # in python3 and later open(<filename>,'rb') creates a binary file stream which has to be 
    # explicitly decoded to unicode strings which are standard in python3 and later. Therefore
    # any regular expression and string used to manipulate the raw stream data also has to be
    # defined as byte string instead of regular raw pyhton string
    _strip_lineend = b'\n'
    _stream_delimiters = [
        re.compile(b"(?:^|\n)@(?P<stream>\\d+)\n", flags=re.S),
        re.compile(r"(?:^|\n)\s*(?P<stream>(?:{}))(?:\s+(?:(?P<count>\d+)|(?P<name>\w+)))?(?:\s*\n|\s+{{)".format(
            _hyper_surface_entities).encode('ASCII')),
        re.compile(b"^\\s*}", re.I)  # NOTE this is applied to reverse slice of stream_data therefore ^
    ]
else:
    # definitions required by python2.x and older which does not destinguish between string and 
    # binary byte string as standard string are still relying on ASCII and alike encoding.
    # Therefore regular stirngs can be used to define all the necessary pattern without encoding
    # them to ASCII byte string
    _file_format_match = (
        re.compile(r'.*AmiraMesh.*'),
        re.compile(r'.*HyperSurface.*'),
        re.compile(r'.*Avizo.*')
    )

    # in python2.x and before strings are per default ascii type strings and thus open(<filename>,'rb') does not 
    # return standarad filesream. Therefore all regulare expression and strings used to manipulate 
    # the raw byte stream can be formulated using regular raw strings
    _strip_lineened = r'\n}{\t '
    _stream_delimiters = [
        re.compile(r"(?:^|\n)@(?P<stream>\d+)\n", flags=re.S),
        re.compile(r"(?:^|\n)\s*(?P<stream>(?:{}))(?:\s+(?:(?P<count>\d+)|(?P<name>\w+)))?(?:\s*\n|\s+{{)".format(
            r"".join(_hyper_surface_entities))),
        re.compile(r"^\s*}'", re.I)  # NOTE this is applied to reverese slice of stream_data therefore ^
    ]


def detect_format(fn, format_bytes=50, verbose=False, *args, **kwargs):
    """Detect Amira (R) file format (AmiraMesh/Avizo or HyperSurface)
    
    :param str fn: file name
    :param int format_bytes: number of bytes in which to search for the format [default: 50]
    :param bool verbose: verbose (default) or not
    :return str file_format: either ``AmiraMesh`` or ``HyperSurface``
    """
    assert format_bytes > 0
    assert verbose in [True, False]

    with open(fn, 'rb') as f:
        rough_header = f.read(format_bytes)

        if _file_format_match[0].match(rough_header):
            file_format = "AmiraMesh"
        elif _file_format_match[1].match(rough_header):
            file_format = "HyperSurface"
        elif _file_format_match[2].match(rough_header):
            file_format = "Avizo"
        else:
            file_format = "Undefined"

    if verbose:
        print("{} file detected...".format(file_format), file=sys.stderr)

    return file_format


SEQ_MAP = [
    (b'\xc5', u'Å'.encode('utf-8')),  # Angstrom char
]


def _swap_illegal_chars(byte_seq, seq_map):
    """Replace illegal byte sequences with legal ones"""
    for s, r in seq_map:
        _byte_seq = byte_seq
        while _byte_seq.find(s) > 1:
            _byte_seq = _byte_seq.replace(s, r)
    swapped_byte_seq = _byte_seq
    return swapped_byte_seq


def get_header(fn, file_format, header_bytes=20000, verbose=False, *args, **kwargs):
    """Apply rules for detecting the boundary of the header
    
    :param str fn: file name
    :param str file_format: either ``AmiraMesh`` or ``HyperSurface``
    :param bool verbose: verbose output; default False
    :param int header_bytes: number of bytes in which to search for the header [default: 20000]
    :return str data: the header as per the ``file_format``
    """
    assert header_bytes > 0
    try:
        assert file_format in ['AmiraMesh', 'HyperSurface', 'Avizo']
    except AssertionError:
        raise ValueError("unknown file format: {}".format(file_format))

    with open(fn, 'rb') as f:
        # read a first chunk and store it in the first element of the list of header chunks
        _data = f.read(header_bytes if header_bytes >= _rescan_overlap else _rescan_overlap)

        data = _swap_illegal_chars(_data, SEQ_MAP)

        if file_format == "AmiraMesh" or file_format == "Avizo":
            if verbose:
                print("Using pattern: {}".format(_stream_delimiters[0].pattern), file=sys.stderr)
            # scan the latests chunk  for the first @<n> data block start marker
            m = _stream_delimiters[0].search(data)
            while m is None:
                _chunklen = len(data) - _rescan_overlap
                data += f.read(header_bytes)
                m = _stream_delimiters[0].search(data, _chunklen)
        elif file_format == "HyperSurface":
            if verbose:
                print("Using pattern: {}".format(_stream_delimiters[1].pattern), file=sys.stderr)
            # scan the latests chunk for the the first occurance of any of the keys of the above
            # _hyper_surface_file structure
            m = _stream_delimiters[1].search(data)
            while m is None:
                _chunklen = len(data) - _rescan_overlap
                data += f.read(header_bytes)
                m = _stream_delimiters[1].search(data, _chunklen)
        elif file_format == "Undefined":
            raise ValueError("Unable to parse undefined file")
    # cut the data before the delimter and encode the remaining byte string into ASCII
    # string in case of python 2 and UTF-8 string for python3
    return _decode_string(data[:m.start()])


def parse_header(data, verbose=False, *args, **kwargs):
    """Parse the data using the grammar specified in this module
    
    :param str data: delimited data to be parsed for metadata
    :param bool verbose: verbose output; default False
    :return list parsed_data: structured metadata
    """
    # the parser
    if verbose:
        print("Creating parser object...", file=sys.stderr)
    parser = Parser(amira_header_grammar)

    # the processor
    if verbose:
        print("Defining dispatch processor...", file=sys.stderr)
    amira_processor = AmiraDispatchProcessor()

    # parsing
    if verbose:
        print("Parsing data...", file=sys.stderr)
    success, parsed_data, next_item = parser.parse(data, production='amira', processor=amira_processor)

    if success:
        if verbose:
            print("Successfully parsed data...", file=sys.stderr)
        return parsed_data
    else:
        raise TypeError("Parse: {}\nNext: {}\n".format(parsed_data, next_item))


def get_parsed_data(fn, *args, **kwargs):
    """All above functions as a single function
    
    :param str fn: file name
    :return tuple(list,int) parsed_data,header_length: structured metadata and total number of header bytes
    """
    file_format = detect_format(fn, *args, **kwargs)
    data = get_header(fn, file_format, *args, **kwargs)
    parsed_data = parse_header(data, *args, **kwargs)
    return data, parsed_data, len(data), file_format



# -*- coding: utf-8 -*-
"""
console entry point
===================

The `ahds` command calls this script which performs the following tasks:

* parse command-line arguments
* run the command

"""

from __future__ import print_function

import argparse
import os
import sys

from . import AmiraFile, WIDTH
from .core import _str


def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(prog='ahds', description='Python tool to read and display Amira files')
    parser.add_argument('file', nargs='+', help='a valid Amira file with an optional block path')
    parser.add_argument('-s', '--load-streams', default=False, action='store_true',
                        help="whether to load data streams or not [default: False]")
    parser.add_argument('-d', '--debug', default=False, action='store_true',
                        help="display debugging information [default: False]")
    parser.add_argument('-l', '--literal', default=False, action='store_true',
                        help="display the literal header [default: False]")

    args = parser.parse_args()
    return args


def main():
    args = parse_args()

    _file, _paths = set_file_and_paths(args)

    af = get_amira_file(_file, args)

    if args.literal:
        print(get_literal(af, args), file=sys.stderr)
    if args.debug:
        print(get_debug(af, args), file=sys.stderr)
    # always show paths
    print(get_paths(_paths, af), file=sys.stderr)
    return os.EX_OK


def get_amira_file(_file, args):
    af = AmiraFile(_file, load_streams=args.load_streams, debug=args.debug)
    return af


def get_paths(_paths, af):
    if _paths:
        string = ""
        for _path in _paths:
            _path_list = _path.split('.')
            current_block = af  # the AmiraFile object
            for block in _path_list:
                current_block = getattr(current_block, block, None)
            if current_block is None:
                print("""Path '{}' not found.""".format(_path))
            else:
                string += u'*' * WIDTH + u'\n'
                string += u"ahds: Displaying path '{}'\n".format(_path)
                string += u"-" * WIDTH + "\n"
                string += _str(current_block)
    else:
        string = _str(af)
    return string


def get_debug(af, args):
    string = ""
    if args.debug:
        from pprint import pformat
        string += u"*" * WIDTH + "\n"
        string += u"ahds: Displaying parsed header data\n"
        string += u"-" * WIDTH + "\n"
        string += pformat(af.header.parsed_data) + '\n'
    return string


def get_literal(af, args):
    string = u""
    if args.literal:
        string += u'*' * WIDTH + u'\n'
        string += u"ahds: Displaying literal header\n"
        string += u"-" * WIDTH + "\n"
        string += _str(af.header.literal_data)
    return string #.encode('utf-8')


def set_file_and_paths(args):
    if len(args.file) == 1:
        _file = args.file[0]
        _paths = None
    else:
        _file = args.file[0]
        _paths = args.file[1:]
    return _file, _paths


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import re
import sys

import ahds
from . import Py23FixTestCase, TEST_DATA_PATH
from ..ahds import parse_args, get_debug, get_literal, get_paths, set_file_and_paths, get_amira_file
from ..core import _str, _print
import numpy

def _parse_with_shlex(cmd):
    import shlex
    import sys
    sys.argv = shlex.split(cmd)
    args = parse_args()
    return args


class TestArgs(Py23FixTestCase):
    """Tests for the main ahds entry point options"""

    def test_file_only(self):
        """Test command with file only"""
        args = _parse_with_shlex("ahds file.am")
        self.assertFalse(args.debug)
        self.assertEqual(args.file, ['file.am'])
        self.assertFalse(args.literal)
        self.assertFalse(args.load_streams)

    def test_file_with_path(self):
        """Test command with file and path"""
        args = _parse_with_shlex("ahds file.am header.Parameters")
        self.assertFalse(args.debug)
        self.assertEqual(args.file, ['file.am', 'header.Parameters'])
        self.assertFalse(args.literal)
        self.assertFalse(args.load_streams)

    def test_debug(self):
        """Test debug option"""
        args = _parse_with_shlex("ahds -d file.am")
        self.assertTrue(args.debug)

    def test_literal(self):
        """Test literal option"""
        args = _parse_with_shlex("ahds -l file.am")
        self.assertTrue(args.literal)

    def test_load_streams(self):
        """Test load_streams option"""
        args = _parse_with_shlex("ahds -s file.am")
        self.assertTrue(args.load_streams)


class TestMain(Py23FixTestCase):
    """Tests for the main ahds entry point main()"""

    @classmethod
    def setUpClass(cls):
        cls.af_fn = os.path.join(TEST_DATA_PATH, 'test12.am')

    def test_set_file_and_paths(self):
        """Test that we correctly extract file and one or more path"""
        # no paths
        args = _parse_with_shlex("ahds {}".format(self.af_fn))
        f, p = set_file_and_paths(args)
        self.assertEqual(f, self.af_fn)
        self.assertEqual(p, None)
        # one path
        args = _parse_with_shlex("ahds {} path1".format(self.af_fn))
        f, p = set_file_and_paths(args)
        self.assertEqual(f, self.af_fn)
        self.assertEqual(p, ['path1'])
        # two paths
        args = _parse_with_shlex("ahds {} path1 path2".format(self.af_fn))
        f, p = set_file_and_paths(args)
        self.assertEqual(f, self.af_fn)
        self.assertEqual(p, ['path1', 'path2'])
        # three paths
        args = _parse_with_shlex("ahds {} path1 path2 path3".format(self.af_fn))
        f, p = set_file_and_paths(args)
        self.assertEqual(f, self.af_fn)
        self.assertEqual(p, ['path1', 'path2', 'path3'])

    def test_get_amira_file(self):
        """Test that we can get the Amira (R) file"""
        args = _parse_with_shlex("ahds -d {}".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        self.assertIsInstance(af, ahds.AmiraFile)

    def test_get_debug(self):
        """Test that we can get debug info"""
        args = _parse_with_shlex("ahds -d {}".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        string = get_debug(af, args)
        self.assertIsInstance(string, _str)
        des = re.compile(r".*\'designation\'.*", re.S)
        com = re.compile(r".*\'comment\'.*", re.S)
        par = re.compile(r".*\'parameters\'.*", re.S)
        dat = re.compile(r".*\'data_definitions\'.*", re.S)
        des_m = des.match(string)
        self.assertIsNotNone(des_m)
        com_m = com.match(string)
        self.assertIsNotNone(com_m)
        par_m = par.match(string)
        self.assertIsNotNone(par_m)
        dat_m = dat.match(string)
        self.assertIsNotNone(dat_m)

    def test_get_literal(self):
        """Test that we can get literal header info"""
        args = _parse_with_shlex("ahds -l {}".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        string = get_literal(af, args)
        self.assertIsInstance(string, _str)
        am = re.compile(r".*AmiraMesh 3D BINARY.*", re.S)
        cd = re.compile(r".*CreationDate.*", re.S)
        de = re.compile(r".*define Lattice.*", re.S)
        par = re.compile(r".*Parameters.*", re.S)
        mat = re.compile(r".*Materials.*", re.S)
        am_m = am.match(string)
        self.assertIsNotNone(am_m)
        cd_m = cd.match(string)
        self.assertIsNotNone(cd_m)
        de_m = de.match(string)
        self.assertIsNotNone(de_m)
        par_m = par.match(string)
        self.assertIsNotNone(par_m)
        mat_m = mat.match(string)
        self.assertIsNotNone(mat_m)

    def test_get_paths_full(self):
        """Test that we can view the paths full"""
        args = _parse_with_shlex("ahds {}".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        string = get_paths(p, af)
        print(string, file=sys.stderr)
        self.assertIsInstance(string, _str)
        am = re.compile(r".*AmiraFile.*", re.S)
        m = re.compile(r".*meta.*", re.S)
        h = re.compile(r".*header.*", re.S)
        ds = re.compile(r".*data_streams.*", re.S)
        am_m = am.match(string)
        self.assertIsNotNone(am_m)
        m_m = m.match(string)
        self.assertIsNotNone(m_m)
        h_m = h.match(string)
        self.assertIsNotNone(h_m)
        ds_m = ds.match(string)
        self.assertIsNotNone(ds_m)

    def test_get_paths_meta(self):
        """Test that we can fiew partial paths"""
        args = _parse_with_shlex("ahds {} meta.streams_loaded".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        string = get_paths(p, af)
        print(string, file=sys.stderr)
        self.assertIsInstance(string, _str)
        am = re.compile(r".*AmiraFile.*", re.S)
        m = re.compile(r".*streams_loaded.*", re.S)
        am_m = am.match(string)
        self.assertIsNone(am_m)
        m_m = m.match(string)
        self.assertIsNotNone(m_m)

    def test_get_paths_header(self):
        """Test that we can fiew partial paths"""
        args = _parse_with_shlex("ahds {} header.Parameters.Materials".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        string = get_paths(p, af)
        self.assertIsInstance(string, _str)
        am = re.compile(r".*AmiraFile.*", re.S)
        m = re.compile(r".*Inside.*", re.S)
        am_m = am.match(string)
        self.assertIsNone(am_m)
        m_m = m.match(string)
        self.assertIsNotNone(m_m)

    def test_get_paths_data_streams(self):
        """Test that we can fiew partial paths"""
        args = _parse_with_shlex("ahds {} data_streams".format(self.af_fn))
        f, p = set_file_and_paths(args)
        af = get_amira_file(f, args)
        string = get_paths(p, af)
        self.assertIsInstance(string, _str)
        am = re.compile(r".*AmiraFile.*", re.S)
        m = re.compile(r".*data_streams.*", re.S)
        am_m = am.match(string)
        self.assertIsNone(am_m)
        m_m = m.match(string)
        self.assertIsNotNone(m_m)

    # def test_data(self):
    #     """Test that the data is correctly oriented"""
    #     af = ahds.AmiraFile(os.path.join(TEST_DATA_PATH, 'EM04226_2_U19_Cropped_YZ_binned.labels.am'))
        # af = ahds.AmiraFile(os.path.join(TEST_DATA_PATH, 'testscalar.am'))
        # _print(af)
        # _print(af.data_streams.Labels.data.shape)
        # _print(af.header.Lattice.length)
        # import h5py
        # with h5py.File('EM04226_2_U19_Cropped_YZ_binned.labels.h5', 'w') as h:
        #     h['/data'] = af.data_streams.Labels.data
        # import mrcfile
        # _print(af.data_streams.Labels.data.dtype)
        # _print('unique values: ', numpy.unique(af.data_streams.Labels.data.astype(numpy.float32)))
        # with mrcfile.new('EM04226_2_U19_Cropped_YZ_binned.labels.mrc', overwrite=True) as m:
        #     m.set_data(af.data_streams.Labels.data.astype(numpy.float32))
        #     m.voxel_size = (9.0, 9.0, 15.0)
        #     m.update_header_from_data()





# -*- coding: utf-8 -*-
# setup.py
from __future__ import print_function

import os
import sys

import numpy as np

AHDS_VERSION = '0.2.4'

from setuptools import setup, find_packages, Extension

"""
The build may fail on macOS. Try the following:

open /Library/Developer/CommandLineTools/Packages/macOS_SDK_headers_for_macOS_10.14.pkg

Credits: https://blog.driftingruby.com/updated-to-mojave/
"""
decoders = Extension(
    'ahds.decoders',
    sources=['src/decodersmodule.cpp'],
)

here = os.path.abspath(os.path.dirname(__file__))

# long description
with open(os.path.join(here, 'README.rst')) as f:
    long_description = f.read()

AHDS_NAME = "ahds"
AHDS_AUTHOR = "Paul K. Korir, PhD"
AHDS_AUTHOR_EMAIL = "pkorir@ebi.ac.uk, paul.korir@gmail.com"
AHDS_DESCRIPTION = "Python package to parse and provide access to headers and data streams in Amira (R) files"
AHDS_DESCRIPTION_CONTENT_TYPE = 'text/x-rst'
AHDS_URL = "https://github.com/emdb-empiar/ahds.git"
AHDS_LICENSE = "Apache License"
AHDS_KEYWORDS = "header, parser, data streams"
AHDS_ENTRY_POINT = 'ahds = ahds.ahds:main'
AHDS_CLASSIFIERS = [
    u"Development Status :: 2 - Pre-Alpha",
    u"Environment :: Console",
    u"Intended Audience :: Developers",
    u"License :: OSI Approved :: Apache Software License",
    u"Operating System :: OS Independent",
    u"Programming Language :: Python :: 2",
    u"Programming Language :: Python :: 2.7",
    u"Programming Language :: Python :: 3",
    u"Programming Language :: Python :: 3.5",
    u"Programming Language :: Python :: 3.6",
    u"Programming Language :: Python :: 3.7",
    u"Topic :: Software Development :: Libraries :: Python Modules",
    u"Topic :: Terminals",
    u"Topic :: Text Processing",
    u"Topic :: Text Processing :: Markup",
    u"Topic :: Utilities",
]

if sys.version_info[0] > 2:
    AHDS_INSTALL_REQUIRES = ["simpleparse"]
    if sys.version_info[1] > 5:
        AHDS_INSTALL_REQUIRES += ["scikit-image"]
    else:
        AHDS_INSTALL_REQUIRES += ["matplotlib<3.1", "scikit-image<0.16"]
    setup(
        name=AHDS_NAME,
        version=AHDS_VERSION,
        packages=find_packages(),
        author=AHDS_AUTHOR,
        author_email=AHDS_AUTHOR_EMAIL,
        description=AHDS_DESCRIPTION,
        long_description=long_description,
        long_description_content_type=AHDS_DESCRIPTION_CONTENT_TYPE,
        url=AHDS_URL,
        license=AHDS_LICENSE,
        keywords=AHDS_KEYWORDS,
        setup_requires=["numpy"],
        # additional dependencies to prevent failed install due to no support for Py27
        install_requires=AHDS_INSTALL_REQUIRES,
        ext_modules=[decoders],
        include_dirs=[np.get_include()],
        entry_points={
            'console_scripts': [
                AHDS_ENTRY_POINT,
            ]
        },
        classifiers=AHDS_CLASSIFIERS,
    )
else:
    setup(
        name=AHDS_NAME,
        version=AHDS_VERSION,
        packages=find_packages(),
        author=AHDS_AUTHOR,
        author_email=AHDS_AUTHOR_EMAIL,
        description=AHDS_DESCRIPTION,
        long_description=long_description,
        long_description_content_type=AHDS_DESCRIPTION_CONTENT_TYPE,
        url=AHDS_URL,
        license=AHDS_LICENSE,
        keywords=AHDS_KEYWORDS,
        setup_requires=["numpy"],
        # additional dependencies to prevent failed install due to no support for Py27
        install_requires=["simpleparse==2.1.1", "scikit-image<0.14", "scipy<1.2", "backports.shutil_get_terminal_size",
                          "networkx==2.2", "matplotlib<3.0", "PyWavelets<1.1.0", "Pillow<7.0.0"],
        ext_modules=[decoders],
        include_dirs=[np.get_include()],
        entry_points={
            'console_scripts': [
                AHDS_ENTRY_POINT,
            ]
        },
        classifiers=AHDS_CLASSIFIERS,
    )



# -*- coding: utf-8 -*-
from __future__ import print_function

import os
import sys
from unittest import TestCase

TEST_DATA_PATH = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'data')


class Py23FixTestCase(TestCase):
    """Mixin to fix method changes in TestCase class"""

    def __init__(self, *args, **kwargs):
        if sys.version_info[0] > 2:
            pass
        else:
            # new names for assert methods
            self.assertCountEqual = self.assertItemsEqual
        super(Py23FixTestCase, self).__init__(*args, **kwargs)
